# Unit tests for speech_presenter.py methods.
#
# Copyright 2025 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=wrong-import-position
# pylint: disable=import-outside-toplevel
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-statements
# pylint: disable=too-many-lines
# pylint: disable=protected-access
# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
# pylint: disable=too-many-locals

"""Unit tests for speech_presenter.py methods."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from unittest.mock import MagicMock

    from .orca_test_context import OrcaTestContext


@pytest.mark.unit
class TestSpeechPresenter:
    """Test SpeechPresenter class methods."""

    def _setup_dependencies(self, test_context: OrcaTestContext) -> dict[str, MagicMock]:
        """Set up mocks for speech_presenter dependencies."""

        additional_modules = [
            "orca.document_presenter",
            "orca.mathsymbols",
            "orca.object_properties",
            "orca.phonnames",
            "orca.pronunciation_dictionary_manager",
            "orca.ax_hypertext",
            "orca.ax_table",
            "orca.ax_text",
            "orca.ax_utilities",
            "orca.ax_document",
            "orca.presentation_manager",
            "orca.preferences_grid_base",
            "orca.speech",
            "orca.speech_manager",
            "orca.speech_monitor",
        ]
        essential_modules = test_context.setup_shared_dependencies(additional_modules)

        document_presenter_mock = essential_modules["orca.document_presenter"]
        document_presenter_instance = test_context.Mock()
        document_presenter_instance.get_in_focus_mode = test_context.Mock(return_value=False)
        document_presenter_mock.get_presenter = test_context.Mock(
            return_value=document_presenter_instance,
        )

        dbus_service_mock = essential_modules["orca.dbus_service"]
        dbus_service_mock.get_remote_controller.return_value = test_context.Mock()

        def passthrough_decorator(func):
            return func

        dbus_service_mock.getter = passthrough_decorator
        dbus_service_mock.setter = passthrough_decorator
        dbus_service_mock.command = passthrough_decorator
        dbus_service_mock.parameterized_command = passthrough_decorator

        debug_mock = essential_modules["orca.debug"]
        debug_mock.print_message.return_value = None
        debug_mock.print_tokens.return_value = None
        debug_mock.LEVEL_INFO = 800
        debug_mock.LEVEL_WARNING = 900

        ax_hypertext_mock = essential_modules["orca.ax_hypertext"]
        ax_hypertext_mock.AXHypertext = test_context.Mock()
        ax_hypertext_mock.AXHypertext.get_all_links_in_range = test_context.Mock(return_value=[])
        ax_hypertext_mock.AXHypertext.get_link_end_offset = test_context.Mock(return_value=0)

        ax_object_mock = essential_modules["orca.ax_object"]
        ax_object_mock.AXObject = test_context.Mock()

        ax_text_mock = essential_modules["orca.ax_text"]
        ax_text_mock.AXText = test_context.Mock()
        ax_text_mock.AXText.get_character_at_offset = test_context.Mock(return_value=("a", 0))

        ax_utilities_mock = essential_modules["orca.ax_utilities"]
        ax_utilities_mock.AXUtilities = test_context.Mock()
        ax_utilities_mock.AXUtilities.find_ancestor_inclusive = test_context.Mock(return_value=None)
        ax_utilities_mock.AXUtilities.string_has_spelling_error = test_context.Mock(
            return_value=False,
        )
        ax_utilities_mock.AXUtilities.string_has_grammar_error = test_context.Mock(
            return_value=False,
        )
        ax_utilities_mock.AXUtilities.get_table = test_context.Mock(return_value=None)
        ax_utilities_mock.AXUtilities.is_math_related = test_context.Mock(return_value=False)
        ax_utilities_mock.AXUtilities.is_text_input_telephone = test_context.Mock(
            return_value=False,
        )
        ax_utilities_mock.AXUtilities.is_code = test_context.Mock(return_value=False)

        ax_document_mock = essential_modules["orca.ax_document"]
        ax_document_mock.AXDocument = test_context.Mock()
        ax_document_mock.AXDocument.is_plain_text = test_context.Mock(return_value=False)

        mathsymbols_mock = essential_modules["orca.mathsymbols"]
        mathsymbols_mock.adjust_for_speech = test_context.Mock(side_effect=lambda x: x)

        object_properties_mock = essential_modules["orca.object_properties"]
        object_properties_mock.STATE_INVALID_GRAMMAR_SPEECH = "grammar error"

        pronunciation_dict_mock = essential_modules["orca.pronunciation_dictionary_manager"]
        pron_manager_instance = test_context.Mock()
        pron_manager_instance.get_pronunciation = test_context.Mock(side_effect=lambda x: x)
        pronunciation_dict_mock.get_manager = test_context.Mock(return_value=pron_manager_instance)

        from orca import gsettings_registry

        gsettings_registry.get_registry().clear_runtime_values()

        orca_i18n_mock = essential_modules["orca.orca_i18n"]
        orca_i18n_mock._ = lambda x: x
        orca_i18n_mock.C_ = lambda c, x: x
        orca_i18n_mock.ngettext = lambda s, p, n: s if n == 1 else p

        messages_mock = essential_modules["orca.messages"]
        messages_mock.LINK = "link"
        messages_mock.MISSPELLED = "misspelled"
        messages_mock.repeated_char_count = test_context.Mock(
            side_effect=lambda char, count: f"{char} repeated {count} times",
        )
        messages_mock.spaces_count = test_context.Mock(side_effect=lambda count: f"{count} spaces")
        messages_mock.tabs_count = test_context.Mock(side_effect=lambda count: f"{count} tabs")

        return essential_modules

    def test_init(self, test_context: OrcaTestContext) -> None:
        """Test presenter initialization and D-Bus registration."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()

        # Verify D-Bus registration occurred
        controller = essential_modules["orca.dbus_service"].get_remote_controller()
        controller.register_decorated_module.assert_called_with("SpeechPresenter", presenter)

        # Monitor callbacks are not registered in __init__ (deferred to set_up_commands
        # to avoid circular imports during module loading).

    def test_set_up_commands(self, test_context: OrcaTestContext) -> None:
        """Test that set_up_commands registers commands in CommandManager."""

        self._setup_dependencies(test_context)
        from orca import command_manager
        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        presenter.set_up_commands()

        # Verify commands are registered in CommandManager
        cmd_manager = command_manager.get_manager()
        assert cmd_manager.get_command("changeNumberStyleHandler") is not None
        assert cmd_manager.get_command("toggleSpeechVerbosityHandler") is not None
        assert cmd_manager.get_command("toggleSpeakingIndentationJustificationHandler") is not None
        assert cmd_manager.get_command("toggleTableCellReadModeHandler") is not None

    def test_set_up_commands_registers_monitor_callbacks(
        self,
        test_context: OrcaTestContext,
    ) -> None:
        """Test that set_up_commands registers speech monitor callbacks."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        speech_mock = essential_modules["orca.speech"]
        speech_mock.set_monitor_callbacks.reset_mock()

        presenter.set_up_commands()

        speech_mock.set_monitor_callbacks.assert_called_once_with(
            write_text=presenter.write_to_monitor,
            write_key=presenter.write_key_to_monitor,
            begin_group=presenter._begin_monitor_group,
            end_group=presenter._end_monitor_group,
        )

    @pytest.mark.parametrize(
        "case",
        [
            {"id": "verbose_true", "setting_value": 1, "expected": True},
            {"id": "verbose_false", "setting_value": 0, "expected": False},
        ],
        ids=lambda case: case["id"],
    )
    def test_use_verbose_speech(self, test_context: OrcaTestContext, case: dict) -> None:
        """Test use_verbose_speech method."""

        self._setup_dependencies(test_context)

        from orca import gsettings_registry
        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        nick = "verbose" if case["setting_value"] == 1 else "brief"
        gsettings_registry.get_registry().set_runtime_value("speech", "verbosity-level", nick)

        result = presenter.use_verbose_speech()
        assert result == case["expected"]

    @pytest.mark.parametrize(
        "case",
        [
            {"id": "verbosity_brief", "setting_value": 0, "expected": "brief"},
            {"id": "verbosity_verbose", "setting_value": 1, "expected": "verbose"},
        ],
        ids=lambda case: case["id"],
    )
    def test_get_verbosity_level(self, test_context: OrcaTestContext, case: dict) -> None:
        """Test get_verbosity_level method."""

        self._setup_dependencies(test_context)

        from orca import gsettings_registry
        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        nick = "verbose" if case["setting_value"] == 1 else "brief"
        gsettings_registry.get_registry().set_runtime_value("speech", "verbosity-level", nick)

        result = presenter.get_verbosity_level()
        assert result == case["expected"]

    @pytest.mark.parametrize(
        "case",
        [
            {
                "id": "set_verbosity_brief",
                "input_value": "brief",
                "expected": True,
            },
            {
                "id": "set_verbosity_verbose",
                "input_value": "verbose",
                "expected": True,
            },
            {
                "id": "set_verbosity_invalid",
                "input_value": "invalid",
                "expected": False,
            },
        ],
        ids=lambda case: case["id"],
    )
    def test_set_verbosity_level(self, test_context: OrcaTestContext, case: dict) -> None:
        """Test set_verbosity_level method."""

        self._setup_dependencies(test_context)

        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()

        result = presenter.set_verbosity_level(case["input_value"])
        assert result == case["expected"]

        if case["expected"]:
            assert presenter.get_verbosity_level() == case["input_value"]

    @pytest.mark.parametrize(
        "case",
        [
            {"id": "speak_blank_lines_true", "setting_value": True, "expected": True},
            {"id": "speak_blank_lines_false", "setting_value": False, "expected": False},
        ],
        ids=lambda case: case["id"],
    )
    def test_get_speak_blank_lines(self, test_context: OrcaTestContext, case: dict) -> None:
        """Test get_speak_blank_lines method."""

        self._setup_dependencies(test_context)

        from orca import gsettings_registry
        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        gsettings_registry.get_registry().set_runtime_value(
            "speech",
            "speak-blank-lines",
            case["setting_value"],
        )

        result = presenter.get_speak_blank_lines()
        assert result == case["expected"]

    @pytest.mark.parametrize(
        "case",
        [
            {"id": "only_speak_displayed_true", "setting_value": True, "expected": True},
            {"id": "only_speak_displayed_false", "setting_value": False, "expected": False},
        ],
        ids=lambda case: case["id"],
    )
    def test_get_only_speak_displayed_text(self, test_context: OrcaTestContext, case: dict) -> None:
        """Test get_only_speak_displayed_text method."""

        self._setup_dependencies(test_context)

        from orca import gsettings_registry
        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        gsettings_registry.get_registry().set_runtime_value(
            "speech",
            "only-speak-displayed-text",
            case["setting_value"],
        )

        result = presenter.get_only_speak_displayed_text()
        assert result == case["expected"]

    @pytest.mark.parametrize(
        "case",
        [
            {"id": "set_only_speak_displayed_true", "input_value": True, "expected": True},
            {"id": "set_only_speak_displayed_false", "input_value": False, "expected": True},
        ],
        ids=lambda case: case["id"],
    )
    def test_set_only_speak_displayed_text(self, test_context: OrcaTestContext, case: dict) -> None:
        """Test set_only_speak_displayed_text method."""

        self._setup_dependencies(test_context)

        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()

        result = presenter.set_only_speak_displayed_text(case["input_value"])
        assert result == case["expected"]
        assert presenter.get_only_speak_displayed_text() == case["input_value"]

    @pytest.mark.parametrize(
        "case",
        [
            {"id": "speak_indentation_true", "setting_value": True, "expected": True},
            {"id": "speak_indentation_false", "setting_value": False, "expected": False},
        ],
        ids=lambda case: case["id"],
    )
    def test_get_speak_indentation_and_justification(
        self,
        test_context: OrcaTestContext,
        case: dict,
    ) -> None:
        """Test get_speak_indentation_and_justification method."""

        self._setup_dependencies(test_context)

        from orca import gsettings_registry
        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        gsettings_registry.get_registry().set_runtime_value(
            "speech",
            "speak-indentation-and-justification",
            case["setting_value"],
        )

        result = presenter.get_speak_indentation_and_justification()
        assert result == case["expected"]

    @pytest.mark.parametrize(
        "case",
        [
            {"id": "set_speak_indentation_true", "input_value": True, "expected": True},
            {"id": "set_speak_indentation_false", "input_value": False, "expected": True},
        ],
        ids=lambda case: case["id"],
    )
    def test_set_speak_indentation_and_justification(
        self,
        test_context: OrcaTestContext,
        case: dict,
    ) -> None:
        """Test set_speak_indentation_and_justification method."""

        self._setup_dependencies(test_context)

        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()

        result = presenter.set_speak_indentation_and_justification(case["input_value"])
        assert result == case["expected"]
        assert presenter.get_speak_indentation_and_justification() == case["input_value"]

    def test_get_indentation_description_disabled(self, test_context: OrcaTestContext) -> None:
        """Test get_indentation_description method when disabled."""

        self._setup_dependencies(test_context)

        from orca import gsettings_registry
        from orca.speech_presenter import SpeechPresenter

        registry = gsettings_registry.get_registry()
        registry.set_runtime_value("speech", "only-speak-displayed-text", True)
        registry.set_runtime_value("speech", "speak-indentation-and-justification", False)

        presenter = SpeechPresenter()
        line = "    Hello world"
        result = presenter.get_indentation_description(line)
        assert result == ""

    def test_get_indentation_description_enabled(self, test_context: OrcaTestContext) -> None:
        """Test get_indentation_description method when enabled."""

        self._setup_dependencies(test_context)

        from orca import gsettings_registry
        from orca.speech_presenter import SpeechPresenter

        registry = gsettings_registry.get_registry()
        registry.set_runtime_value("speech", "only-speak-displayed-text", False)
        registry.set_runtime_value("speech", "speak-indentation-and-justification", True)
        registry.set_runtime_value("speech", "speak-indentation-only-if-changed", False)

        presenter = SpeechPresenter()
        line = "    Hello world"
        result = presenter.get_indentation_description(line)
        assert result != ""

    def test_get_indentation_description_only_if_changed(
        self,
        test_context: OrcaTestContext,
    ) -> None:
        """Test get_indentation_description with only-if-changed enabled."""

        self._setup_dependencies(test_context)

        from orca import gsettings_registry
        from orca.speech_presenter import SpeechPresenter

        registry = gsettings_registry.get_registry()
        registry.set_runtime_value("speech", "only-speak-displayed-text", False)
        registry.set_runtime_value("speech", "speak-indentation-and-justification", True)
        registry.set_runtime_value("speech", "speak-indentation-only-if-changed", True)

        presenter = SpeechPresenter()
        line = "    Hello world"

        # First call should return a description
        result1 = presenter.get_indentation_description(line)
        assert result1 != ""

        # Second call with same indentation should return empty
        result2 = presenter.get_indentation_description(line)
        assert result2 == ""

    def test_get_error_description_basic(self, test_context: OrcaTestContext) -> None:
        """Test get_error_description method with basic scenarios."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)

        ax_text_mock = essential_modules["orca.ax_text"]
        ax_text_mock.AXText.get_character_at_offset = test_context.Mock(return_value=("a", 0))
        ax_utilities_mock = essential_modules["orca.ax_utilities"]
        ax_utilities_mock.AXUtilities.string_has_spelling_error = test_context.Mock(
            return_value=True,
        )
        ax_utilities_mock.AXUtilities.string_has_grammar_error = test_context.Mock(
            return_value=False,
        )

        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        mock_obj = test_context.Mock()
        result = presenter.get_error_description(mock_obj, 0)
        assert result == "misspelled"

    def test_get_error_description_disabled(self, test_context: OrcaTestContext) -> None:
        """Test get_error_description method when misspelled indicator is disabled."""

        self._setup_dependencies(test_context)

        from orca import gsettings_registry
        from orca.speech_presenter import SpeechPresenter

        gsettings_registry.get_registry().set_runtime_value(
            "speech",
            "speak-misspelled-indicator",
            False,
        )

        presenter = SpeechPresenter()
        mock_obj = test_context.Mock()
        result = presenter.get_error_description(mock_obj, 0)
        assert result == ""

    def test_adjust_for_presentation(self, test_context: OrcaTestContext) -> None:
        """Test adjust_for_presentation with all sub-adjustments mocked."""

        self._setup_dependencies(test_context)
        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        return_text = "Hello world"

        test_context.patch_object(presenter, "_adjust_for_links", return_value=return_text)
        test_context.patch_object(presenter, "adjust_for_digits", return_value=return_text)
        test_context.patch_object(presenter, "_adjust_for_repeats", return_value=return_text)
        test_context.patch_object(
            presenter,
            "_adjust_for_verbalized_punctuation",
            return_value=return_text,
        )
        test_context.patch_object(
            presenter,
            "_apply_pronunciation_dictionary",
            return_value=return_text,
        )

        mock_obj = test_context.Mock()
        result = presenter.adjust_for_presentation(mock_obj, "Hello world", start_offset=0)
        assert result == return_text

    @pytest.mark.parametrize(
        "case",
        [
            {
                "id": "digits_on",
                "speak_digits": True,
                "is_telephone": False,
                "input_text": "123 Main",
                "expected": "1 2 3 Main",
            },
            {
                "id": "digits_off",
                "speak_digits": False,
                "is_telephone": False,
                "input_text": "123 Main",
                "expected": "123 Main",
            },
            {
                "id": "telephone_field",
                "speak_digits": False,
                "is_telephone": True,
                "input_text": "555",
                "expected": "5 5 5",
            },
        ],
        ids=lambda case: case["id"],
    )
    def test_adjust_for_digits(self, test_context: OrcaTestContext, case: dict) -> None:
        """Test adjust_for_digits method with speakNumbersAsDigits on/off."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)

        from orca import gsettings_registry

        gsettings_registry.get_registry().set_runtime_value(
            "speech",
            "speak-numbers-as-digits",
            case["speak_digits"],
        )

        essential_modules[
            "orca.ax_utilities"
        ].AXUtilities.is_text_input_telephone = test_context.Mock(return_value=case["is_telephone"])

        from orca.speech_presenter import SpeechPresenter

        mock_obj = test_context.Mock()
        result = SpeechPresenter.adjust_for_digits(mock_obj, case["input_text"])
        assert result == case["expected"]

    def test_adjust_for_repeats(self, test_context: OrcaTestContext) -> None:
        """Test _adjust_for_repeats with repeated characters."""

        self._setup_dependencies(test_context)

        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        text = "----"
        result = presenter._adjust_for_repeats(text)
        # Should describe repeated characters
        assert "repeated" in result or result != text

    def test_adjust_for_repeats_short_text(self, test_context: OrcaTestContext) -> None:
        """Test _adjust_for_repeats with text shorter than limit."""

        self._setup_dependencies(test_context)

        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        text = "hi"
        result = presenter._adjust_for_repeats(text)
        # Short text should be returned unchanged
        assert result == text

    def test_get_speech_preferences(self, test_context: OrcaTestContext) -> None:
        """Test get_speech_preferences returns correct tuple structure."""

        self._setup_dependencies(test_context)
        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        result = presenter.get_speech_preferences()

        assert isinstance(result, tuple)
        assert len(result) == 3

        general, object_details, announcements = result

        # general should have 1 preference
        assert len(general) == 1
        assert general[0].prefs_key == "messages-are-detailed"

        # object_details should have 5 preferences
        assert len(object_details) == 5
        assert object_details[0].prefs_key == "only-speak-displayed-text"

        # announcements should have 6 preferences
        assert len(announcements) == 6
        assert announcements[0].prefs_key == "announce-blockquote"

    def test_apply_speech_preferences(self, test_context: OrcaTestContext) -> None:
        """Test apply_speech_preferences applies values correctly."""

        self._setup_dependencies(test_context)
        from orca.speech_presenter import SpeechPreference, SpeechPresenter

        presenter = SpeechPresenter()

        mock_setter1 = test_context.Mock(return_value=True)
        mock_setter2 = test_context.Mock(return_value=True)
        pref1 = SpeechPreference("key1", "Label 1", lambda: True, mock_setter1)
        pref2 = SpeechPreference("key2", "Label 2", lambda: False, mock_setter2)

        updates = [(pref1, False), (pref2, True)]
        result = presenter.apply_speech_preferences(updates)

        assert result == {"key1": False, "key2": True}
        mock_setter1.assert_called_once_with(False)
        mock_setter2.assert_called_once_with(True)

    def test_toggle_indentation_and_justification(self, test_context: OrcaTestContext) -> None:
        """Test toggle_indentation_and_justification method."""

        self._setup_dependencies(test_context)
        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        result = presenter.toggle_indentation_and_justification()
        assert result is True

    def test_change_number_style(self, test_context: OrcaTestContext) -> None:
        """Test change_number_style method."""

        self._setup_dependencies(test_context)
        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        result = presenter.change_number_style()
        assert result is True

    def test_should_verbalize_punctuation_false(self, test_context: OrcaTestContext) -> None:
        """Test _should_verbalize_punctuation static method returns False."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.speech_presenter import SpeechPresenter

        essential_modules[
            "orca.ax_utilities"
        ].AXUtilities.find_ancestor_inclusive.return_value = None
        mock_obj = test_context.Mock()
        result = SpeechPresenter._should_verbalize_punctuation(mock_obj)
        assert result is False

    def test_should_verbalize_punctuation_true(self, test_context: OrcaTestContext) -> None:
        """Test _should_verbalize_punctuation static method returns True."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.speech_presenter import SpeechPresenter

        mock_code_obj = test_context.Mock()
        (
            essential_modules["orca.ax_utilities"].AXUtilities.find_ancestor_inclusive.return_value
        ) = mock_code_obj
        essential_modules["orca.ax_document"].AXDocument.is_plain_text.return_value = False
        mock_obj = test_context.Mock()
        result = SpeechPresenter._should_verbalize_punctuation(mock_obj)
        assert result is True

    def test_adjust_for_verbalized_punctuation(self, test_context: OrcaTestContext) -> None:
        """Test _adjust_for_verbalized_punctuation static method."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.speech_presenter import SpeechPresenter

        mock_code_obj = test_context.Mock()
        (
            essential_modules["orca.ax_utilities"].AXUtilities.find_ancestor_inclusive.return_value
        ) = mock_code_obj
        essential_modules["orca.ax_document"].AXDocument.is_plain_text.return_value = False
        mock_obj = test_context.Mock()
        text = "Hello, world! How are you?"
        result = SpeechPresenter._adjust_for_verbalized_punctuation(mock_obj, text)

        expected = "Hello ,  world !  How are you ? "
        assert result == expected

    @pytest.mark.parametrize(
        "case",
        [
            {"id": "set_speak_blank_lines_true", "input_value": True, "expected": True},
            {"id": "set_speak_blank_lines_false", "input_value": False, "expected": True},
        ],
        ids=lambda case: case["id"],
    )
    def test_set_speak_blank_lines(self, test_context: OrcaTestContext, case: dict) -> None:
        """Test set_speak_blank_lines method."""

        self._setup_dependencies(test_context)

        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()

        result = presenter.set_speak_blank_lines(case["input_value"])
        assert result == case["expected"]
        assert presenter.get_speak_blank_lines() == case["input_value"]

    def test_get_presenter_singleton(self, test_context: OrcaTestContext) -> None:
        """Test get_presenter function returns the same instance."""

        self._setup_dependencies(test_context)
        from orca import speech_presenter

        presenter1 = speech_presenter.get_presenter()
        presenter2 = speech_presenter.get_presenter()

        assert presenter1 is presenter2
        assert isinstance(presenter1, speech_presenter.SpeechPresenter)

    def _setup_speech_output_dependencies(
        self,
        test_context: OrcaTestContext,
    ) -> dict[str, MagicMock]:
        """Set up additional mocks needed for speech output method testing."""

        essential_modules = self._setup_dependencies(test_context)

        # Add speech module mock
        speech_mock = essential_modules["orca.speech"]
        speech_mock.speak = test_context.Mock()
        speech_mock.speak_character = test_context.Mock()
        speech_mock.speak_key_event = test_context.Mock()

        # Add phonnames module mock
        phonnames_mock = essential_modules["orca.phonnames"]
        phonnames_mock.get_phonetic_name = test_context.Mock(side_effect=lambda c: f"phonetic_{c}")

        # Add speech_manager mock
        speech_manager_mock = essential_modules["orca.speech_manager"]
        speech_manager_instance = test_context.Mock()
        speech_manager_instance.get_speech_is_muted = test_context.Mock(return_value=False)
        speech_manager_instance.get_speech_is_enabled_and_not_muted = test_context.Mock(
            return_value=True,
        )
        speech_manager_instance.get_capitalization_style = test_context.Mock(return_value="icon")
        speech_manager_instance.set_capitalization_style = test_context.Mock()
        speech_manager_instance.get_punctuation_level = test_context.Mock(return_value="all")
        speech_manager_instance.set_punctuation_level = test_context.Mock()
        speech_manager_mock.get_manager = test_context.Mock(return_value=speech_manager_instance)

        # Add script_manager mock for _get_active_script / _get_voice
        script_manager_mock = essential_modules["orca.script_manager"]
        mock_script = test_context.Mock()
        speech_gen = test_context.Mock()
        speech_gen.voice = test_context.Mock(return_value=[{"family": "default"}])
        speech_gen.generate_contents = test_context.Mock(return_value=["generated speech"])
        mock_script.get_speech_generator = test_context.Mock(return_value=speech_gen)
        script_manager_instance = test_context.Mock()
        script_manager_instance.get_active_script = test_context.Mock(return_value=mock_script)
        script_manager_mock.get_manager = test_context.Mock(return_value=script_manager_instance)

        return essential_modules

    def test_get_voice_with_active_script(self, test_context: OrcaTestContext) -> None:
        """Test _get_voice returns voice from active script."""

        essential_modules = self._setup_speech_output_dependencies(test_context)
        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        voice = presenter._get_voice(text="test")

        script_manager = essential_modules["orca.script_manager"].get_manager()
        script = script_manager.get_active_script()
        script.get_speech_generator().voice.assert_called_once()
        kwargs = script.get_speech_generator().voice.call_args.kwargs
        assert kwargs["obj"] is None
        assert kwargs["string"] == "test"
        assert kwargs["context"].in_preferences_window is False
        assert voice == [{"family": "default"}]

    def test_get_voice_no_active_script(self, test_context: OrcaTestContext) -> None:
        """Test _get_voice returns empty list when no active script."""

        essential_modules = self._setup_speech_output_dependencies(test_context)
        script_manager = essential_modules["orca.script_manager"].get_manager()
        script_manager.get_active_script.return_value = None

        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        voice = presenter._get_voice(text="test")

        assert voice == []

    def test_speak_message(self, test_context: OrcaTestContext) -> None:
        """Test speak_message speaks text via speech module."""

        essential_modules = self._setup_speech_output_dependencies(test_context)
        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        presenter.speak_message("Hello world")

        essential_modules["orca.speech"].speak.assert_called()

    def test_speak_message_non_string(self, test_context: OrcaTestContext) -> None:
        """Test speak_message with non-string returns early."""

        essential_modules = self._setup_speech_output_dependencies(test_context)
        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        presenter.speak_message(123)  # type: ignore

        essential_modules["orca.debug"].print_exception.assert_called()
        essential_modules["orca.speech"].speak.assert_not_called()

    def test_speak_message_only_displayed_text(self, test_context: OrcaTestContext) -> None:
        """Test speak_message when only_speak_displayed_text is true."""

        essential_modules = self._setup_speech_output_dependencies(test_context)

        from orca import gsettings_registry
        from orca.speech_presenter import SpeechPresenter

        gsettings_registry.get_registry().set_runtime_value(
            "speech",
            "only-speak-displayed-text",
            True,
        )

        presenter = SpeechPresenter()
        presenter.speak_message("Hello world")

        essential_modules["orca.speech"].speak.assert_not_called()

    def test_speak_character(self, test_context: OrcaTestContext) -> None:
        """Test speak_character speaks a single character."""

        essential_modules = self._setup_speech_output_dependencies(test_context)
        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        presenter.speak_character("a")

        essential_modules["orca.speech"].speak_character.assert_called_once()

    def test_spell_item(self, test_context: OrcaTestContext) -> None:
        """Test spell_item speaks each character."""

        essential_modules = self._setup_speech_output_dependencies(test_context)
        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        presenter.spell_item("abc")

        assert essential_modules["orca.speech"].speak_character.call_count == 3

    def test_spell_phonetically(self, test_context: OrcaTestContext) -> None:
        """Test spell_phonetically speaks phonetic names."""

        essential_modules = self._setup_speech_output_dependencies(test_context)
        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        presenter.spell_phonetically("ab")

        essential_modules["orca.phonnames"].get_phonetic_name.assert_called()
        assert essential_modules["orca.speech"].speak.call_count >= 2

    def test_speak_contents(self, test_context: OrcaTestContext) -> None:
        """Test speak_contents generates and speaks contents."""

        essential_modules = self._setup_speech_output_dependencies(test_context)
        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        mock_contents = [(test_context.Mock(), 0, 10, "test text")]
        presenter.speak_contents(mock_contents)

        script_manager = essential_modules["orca.script_manager"].get_manager()
        script = script_manager.get_active_script()
        script.get_speech_generator().generate_contents.assert_called_once()
        essential_modules["orca.speech"].speak.assert_called()

    def test_speak_contents_no_active_script(self, test_context: OrcaTestContext) -> None:
        """Test speak_contents returns early when no active script."""

        essential_modules = self._setup_speech_output_dependencies(test_context)
        script_manager = essential_modules["orca.script_manager"].get_manager()
        script_manager.get_active_script.return_value = None

        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        mock_contents = [(test_context.Mock(), 0, 10, "test text")]
        presenter.speak_contents(mock_contents)

        essential_modules["orca.speech"].speak.assert_not_called()

    def test_present_key_event(self, test_context: OrcaTestContext) -> None:
        """Test present_key_event speaks key via speech module."""

        essential_modules = self._setup_speech_output_dependencies(test_context)
        from orca.speech_presenter import SpeechPresenter

        presenter = SpeechPresenter()
        mock_event = test_context.Mock()
        mock_event.is_printable_key.return_value = True
        mock_event.get_key_name.return_value = "a"
        presenter.present_key_event(mock_event)

        essential_modules["orca.speech"].speak_key_event.assert_called_once()

    def test_get_set_monitor_is_enabled(self, test_context: OrcaTestContext) -> None:
        """Test getting and setting speech monitor enabled status."""

        self._setup_dependencies(test_context)
        from orca.speech_presenter import get_presenter

        presenter = get_presenter()

        result = presenter.set_monitor_is_enabled(True)
        assert result is True
        assert presenter.get_monitor_is_enabled() is True

        result = presenter.set_monitor_is_enabled(False)
        assert result is True
        assert presenter.get_monitor_is_enabled() is False

    def test_ensure_monitor_creates_when_enabled(self, test_context: OrcaTestContext) -> None:
        """Test _ensure_monitor creates monitor on demand when enabled."""

        essential_modules = self._setup_dependencies(test_context)
        speech_monitor_mock = essential_modules["orca.speech_monitor"]
        from orca.speech_presenter import get_presenter

        presenter = get_presenter()
        presenter.set_monitor_is_enabled(True)
        mock_monitor = test_context.Mock()
        speech_monitor_mock.SpeechMonitor.return_value = mock_monitor

        result = presenter._ensure_monitor()

        speech_monitor_mock.SpeechMonitor.assert_called_once()
        mock_monitor.show_all.assert_called_once()
        assert result is mock_monitor

    def test_ensure_monitor_returns_none_when_disabled(self, test_context: OrcaTestContext) -> None:
        """Test _ensure_monitor returns None when disabled."""

        self._setup_dependencies(test_context)
        from orca.speech_presenter import get_presenter

        presenter = get_presenter()
        presenter.set_monitor_is_enabled(False)
        mock_monitor = test_context.Mock()
        presenter._monitor = mock_monitor

        result = presenter._ensure_monitor()

        assert result is None

    def test_write_to_monitor(self, test_context: OrcaTestContext) -> None:
        """Test write_to_monitor writes text when monitor active and not focused."""

        essential_modules = self._setup_dependencies(test_context)
        speech_monitor_mock = essential_modules["orca.speech_monitor"]
        from orca.speech_presenter import get_presenter

        presenter = get_presenter()
        presenter.set_monitor_is_enabled(True)
        mock_monitor = test_context.Mock()
        mock_monitor.has_toplevel_focus.return_value = False
        speech_monitor_mock.SpeechMonitor.return_value = mock_monitor

        presenter.write_to_monitor("hello world")

        mock_monitor.write_text.assert_called_once_with("hello world")

    def test_write_to_monitor_skips_when_focused(self, test_context: OrcaTestContext) -> None:
        """Test write_to_monitor skips when monitor has focus."""

        essential_modules = self._setup_dependencies(test_context)
        speech_monitor_mock = essential_modules["orca.speech_monitor"]
        from orca.speech_presenter import get_presenter

        presenter = get_presenter()
        presenter.set_monitor_is_enabled(True)
        mock_monitor = test_context.Mock()
        mock_monitor.has_toplevel_focus.return_value = True
        speech_monitor_mock.SpeechMonitor.return_value = mock_monitor

        presenter.write_to_monitor("hello world")

        mock_monitor.write_text.assert_not_called()

    def test_write_key_to_monitor(self, test_context: OrcaTestContext) -> None:
        """Test write_key_to_monitor writes key event."""

        essential_modules = self._setup_dependencies(test_context)
        speech_monitor_mock = essential_modules["orca.speech_monitor"]
        from orca.speech_presenter import get_presenter

        presenter = get_presenter()
        presenter.set_monitor_is_enabled(True)
        mock_monitor = test_context.Mock()
        mock_monitor.has_toplevel_focus.return_value = False
        speech_monitor_mock.SpeechMonitor.return_value = mock_monitor

        presenter.write_key_to_monitor("Return")

        mock_monitor.write_key_event.assert_called_once_with("Return")

    def test_destroy_monitor(self, test_context: OrcaTestContext) -> None:
        """Test destroy_monitor destroys existing speech monitor."""

        self._setup_dependencies(test_context)
        from orca.speech_presenter import get_presenter

        presenter = get_presenter()
        mock_monitor = test_context.Mock()
        presenter._monitor = mock_monitor

        presenter.destroy_monitor()

        mock_monitor.destroy.assert_called_once()
        assert presenter._monitor is None

    def test_destroy_monitor_no_op_when_none(self, test_context: OrcaTestContext) -> None:
        """Test destroy_monitor does nothing when no monitor exists."""

        self._setup_dependencies(test_context)
        from orca.speech_presenter import get_presenter

        presenter = get_presenter()
        assert presenter._monitor is None

        presenter.destroy_monitor()

        assert presenter._monitor is None



# Unit tests for bypass_mode_manager.py methods.
#
# Copyright 2025 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-public-methods
# pylint: disable=too-many-locals
# pylint: disable=import-outside-toplevel
# pylint: disable=protected-access
# pylint: disable=no-member

"""Unit tests for bypass_mode_manager.py methods."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from unittest.mock import MagicMock

    from .orca_test_context import OrcaTestContext


@pytest.mark.unit
class TestBypassModeManager:
    """Test BypassModeManager class methods."""

    def _setup_dependencies(self, test_context: OrcaTestContext) -> dict[str, MagicMock]:
        """Returns dependencies for bypass_mode_manager module testing."""

        additional_modules = [
            "orca.orca_modifier_manager",
            "orca.command_manager",
            "orca.braille_presenter",
            "orca.presentation_manager",
        ]
        essential_modules = test_context.setup_shared_dependencies(additional_modules)

        debug_mock = essential_modules["orca.debug"]
        debug_mock.print_message = test_context.Mock()
        debug_mock.LEVEL_INFO = 800

        cmdnames_mock = essential_modules["orca.cmdnames"]
        cmdnames_mock.BYPASS_MODE_TOGGLE = "Toggle all Orca command keys"

        input_event_mock = essential_modules["orca.input_event"]
        input_event_handler_mock = test_context.Mock()
        input_event_mock.InputEventHandler = test_context.Mock(
            return_value=input_event_handler_mock,
        )

        keybindings_mock = essential_modules["orca.keybindings"]
        key_bindings_instance = test_context.Mock()
        key_bindings_instance.is_empty = test_context.Mock(return_value=False)
        key_bindings_instance.key_bindings = []
        key_bindings_instance.add = test_context.Mock(return_value=None)
        keybindings_mock.KeyBindings = test_context.Mock(return_value=key_bindings_instance)
        keybindings_mock.KeyBinding = test_context.Mock()
        keybindings_mock.DEFAULT_MODIFIER_MASK = 1
        keybindings_mock.ALT_MODIFIER_MASK = 2

        messages_mock = essential_modules["orca.messages"]
        messages_mock.BYPASS_MODE_ENABLED = "Orca command keys off."
        messages_mock.BYPASS_MODE_DISABLED = "Orca command keys on."

        orca_modifier_manager_mock = essential_modules["orca.orca_modifier_manager"]
        modifier_manager_instance = test_context.Mock()
        modifier_manager_instance.unset_orca_modifiers = test_context.Mock()
        modifier_manager_instance.refresh_orca_modifiers = test_context.Mock()
        modifier_manager_instance.add_grabs_for_orca_modifiers = test_context.Mock()
        modifier_manager_instance.remove_grabs_for_orca_modifiers = test_context.Mock()
        orca_modifier_manager_mock.get_manager = test_context.Mock(
            return_value=modifier_manager_instance,
        )

        command_manager_mock = essential_modules["orca.command_manager"]
        command_manager_instance = test_context.Mock()
        command_manager_instance.set_all_suspended = test_context.Mock()
        command_manager_instance.get_command = test_context.Mock(return_value=None)
        command_manager_mock.get_manager = test_context.Mock(return_value=command_manager_instance)

        essential_modules["input_event_handler"] = input_event_handler_mock
        essential_modules["key_bindings_instance"] = key_bindings_instance
        essential_modules["modifier_manager_instance"] = modifier_manager_instance
        essential_modules["command_manager_instance"] = command_manager_instance

        return essential_modules

    def test_init(self, test_context: OrcaTestContext) -> None:
        """Test BypassModeManager initialization."""

        self._setup_dependencies(test_context)
        from orca import command_manager
        from orca.bypass_mode_manager import BypassModeManager

        manager = BypassModeManager()
        assert manager._is_active is False
        # Commands are registered during setup(), not __init__()
        manager.set_up_commands()
        cmd_manager = command_manager.get_manager()
        assert cmd_manager.get_keyboard_command("bypass_mode_toggle") is not None

    def test_setup_handlers(self, test_context: OrcaTestContext) -> None:
        """Test that commands are registered with CommandManager during setup."""

        self._setup_dependencies(test_context)
        from orca import command_manager
        from orca.bypass_mode_manager import BypassModeManager

        manager = BypassModeManager()
        manager.set_up_commands()
        # Commands are now registered with CommandManager during setup()
        cmd_manager = command_manager.get_manager()
        cmd = cmd_manager.get_keyboard_command("bypass_mode_toggle")
        assert cmd is not None

    def test_setup_bindings(self, test_context: OrcaTestContext) -> None:
        """Test that keybindings are created via Command.set_keybinding."""

        self._setup_dependencies(test_context)
        from orca import command_manager
        from orca.bypass_mode_manager import BypassModeManager

        manager = BypassModeManager()
        manager.set_up_commands()
        # Verify the command has a keybinding
        cmd_manager = command_manager.get_manager()
        cmd = cmd_manager.get_keyboard_command("bypass_mode_toggle")
        assert cmd is not None
        keybinding = cmd.get_keybinding()
        assert keybinding is not None

    @pytest.mark.parametrize(
        "initial_active,has_event,expected_active,expected_message,expected_modifier_call",
        [
            (
                False,
                True,
                True,
                "Orca command keys off.",
                "unset_orca_modifiers",
            ),
            (
                False,
                False,
                True,
                None,
                "unset_orca_modifiers",
            ),
            (
                True,
                True,
                False,
                "Orca command keys on.",
                "refresh_orca_modifiers",
            ),
            (
                True,
                False,
                False,
                None,
                "refresh_orca_modifiers",
            ),
        ],
    )
    def test_toggle_enabled(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self,
        test_context: OrcaTestContext,
        initial_active: bool,
        has_event: bool,
        expected_active: bool,
        expected_message: str | None,
        expected_modifier_call: str,
    ) -> None:
        """Test toggle_enabled with various initial states and event conditions."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)

        if not initial_active:
            mock_binding1 = test_context.Mock()
            mock_binding2 = test_context.Mock()
            key_bindings = [mock_binding1, mock_binding2]
        else:
            key_bindings = []

        key_bindings_instance = essential_modules["key_bindings_instance"]
        key_bindings_instance.key_bindings = key_bindings
        from orca.bypass_mode_manager import BypassModeManager

        manager = BypassModeManager()
        manager._is_active = initial_active
        mock_script = test_context.Mock()
        mock_event = test_context.Mock() if has_event else None

        result = manager.toggle_enabled(mock_script, mock_event)
        assert result is True
        assert manager._is_active is expected_active

        pres_manager = essential_modules["orca.presentation_manager"].get_manager()
        if expected_message:
            pres_manager.present_message.assert_called_with(expected_message)
        else:
            pres_manager.present_message.assert_not_called()

        cmd_manager = essential_modules["command_manager_instance"]
        cmd_manager.set_all_suspended.assert_called()
        call_args = cmd_manager.set_all_suspended.call_args
        if expected_active:
            assert call_args[0][0] is True
            assert call_args[0][1] == frozenset({"bypass_mode_toggle"})
        else:
            assert call_args[0][0] is False

        modifier_manager = essential_modules["modifier_manager_instance"]
        modifier_method = getattr(modifier_manager, expected_modifier_call)
        if expected_active:
            modifier_method.assert_called_with("bypass mode enabled")
        else:
            modifier_method.assert_called_with("bypass mode disabled")

    def test_toggle_enabled_multiple_times(self, test_context: OrcaTestContext) -> None:
        """Test toggle_enabled works correctly when called multiple times."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)

        key_bindings_instance = essential_modules["key_bindings_instance"]
        key_bindings_instance.key_bindings = []
        from orca.bypass_mode_manager import BypassModeManager

        manager = BypassModeManager()
        mock_script = test_context.Mock()

        assert manager.is_active() is False

        result1 = manager.toggle_enabled(mock_script, None)
        assert result1 is True
        assert manager.is_active() is True

        result2 = manager.toggle_enabled(mock_script, None)
        assert result2 is True
        assert manager.is_active() is False

        result3 = manager.toggle_enabled(mock_script, None)
        assert result3 is True
        assert manager.is_active() is True

    def test_bindings_created_correctly(self, test_context: OrcaTestContext) -> None:
        """Test that bindings are properly configured from commands."""

        self._setup_dependencies(test_context)
        from orca import command_manager
        from orca.bypass_mode_manager import BypassModeManager

        manager = BypassModeManager()
        manager.set_up_commands()
        # Verify bypass_mode_toggle command has keybinding set
        cmd = command_manager.get_manager().get_keyboard_command("bypass_mode_toggle")
        assert cmd is not None
        assert cmd.get_keybinding() is not None

    def test_toggle_enabled_with_minimal_script_interface(
        self,
        test_context: OrcaTestContext,
    ) -> None:
        """Test toggle_enabled method with script that has minimal interface."""

        self._setup_dependencies(test_context)
        from orca.bypass_mode_manager import BypassModeManager

        manager = BypassModeManager()
        mock_script = test_context.Mock(spec=[])
        mock_script.add_key_grabs = test_context.Mock()
        mock_script.remove_key_grabs = test_context.Mock()
        mock_script.present_message = test_context.Mock()
        mock_script.key_bindings = test_context.Mock()
        mock_script.key_bindings.add = test_context.Mock()

        result = manager.toggle_enabled(mock_script)
        assert result is True
        assert manager.is_active() is True



# Unit tests for caret_navigator.py methods.
#
# Copyright 2025 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=wrong-import-position
# pylint: disable=import-outside-toplevel
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-statements
# pylint: disable=protected-access
# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
# pylint: disable=too-many-locals

"""Unit tests for caret_navigator.py methods."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pytest

if TYPE_CHECKING:
    from .orca_test_context import OrcaTestContext


@pytest.mark.unit
class TestCaretNavigator:
    """Test CaretNavigator class methods."""

    def _setup_dependencies(self, test_context: OrcaTestContext) -> dict[str, Any]:
        """Set up mocks for caret_navigator dependencies."""

        additional_modules = [
            "orca.command_manager",
            "orca.input_event_manager",
            "orca.keybindings",
            "orca.cmdnames",
            "orca.guilabels",
            "orca.debug",
            "orca.ax_object",
            "orca.ax_text",
            "orca.script_manager",
            "orca.messages",
            "orca.object_properties",
            "orca.orca_gui_navlist",
            "orca.orca_i18n",
            "orca.AXHypertext",
            "orca.AXObject",
            "orca.AXTable",
            "orca.AXText",
            "orca.AXUtilities",
            "orca.input_event",
            "orca.braille_presenter",
            "orca.presentation_manager",
        ]
        essential_modules = test_context.setup_shared_dependencies(additional_modules)

        # Set up cmdnames with all required values for structural_navigator
        cmdnames = essential_modules["orca.cmdnames"]
        cmdnames.STRUCTURAL_NAVIGATION_MODE_CYCLE = "cycle_mode"
        cmdnames.BLOCKQUOTE_PREV = "previous_blockquote"
        cmdnames.BLOCKQUOTE_NEXT = "next_blockquote"
        cmdnames.BLOCKQUOTE_LIST = "list_blockquotes"
        cmdnames.BUTTON_PREV = "previous_button"
        cmdnames.BUTTON_NEXT = "next_button"
        cmdnames.BUTTON_LIST = "list_buttons"
        cmdnames.CHECK_BOX_PREV = "previous_checkbox"
        cmdnames.CHECK_BOX_NEXT = "next_checkbox"
        cmdnames.CHECK_BOX_LIST = "list_checkboxes"
        cmdnames.COMBO_BOX_PREV = "previous_combobox"
        cmdnames.COMBO_BOX_NEXT = "next_combobox"
        cmdnames.COMBO_BOX_LIST = "list_comboboxes"
        cmdnames.ENTRY_PREV = "previous_entry"
        cmdnames.ENTRY_NEXT = "next_entry"
        cmdnames.ENTRY_LIST = "list_entries"
        cmdnames.FORM_FIELD_PREV = "previous_form_field"
        cmdnames.FORM_FIELD_NEXT = "next_form_field"
        cmdnames.FORM_FIELD_LIST = "list_form_fields"
        cmdnames.HEADING_PREV = "previous_heading"
        cmdnames.HEADING_NEXT = "next_heading"
        cmdnames.HEADING_LIST = "list_headings"
        cmdnames.HEADING_AT_LEVEL_PREV = "previous_heading_level_%d"
        cmdnames.HEADING_AT_LEVEL_NEXT = "next_heading_level_%d"
        cmdnames.HEADING_AT_LEVEL_LIST = "list_headings_level_%d"
        cmdnames.IFRAME_PREV = "previous_iframe"
        cmdnames.IFRAME_NEXT = "next_iframe"
        cmdnames.IFRAME_LIST = "list_iframes"
        cmdnames.IMAGE_PREV = "previous_image"
        cmdnames.IMAGE_NEXT = "next_image"
        cmdnames.IMAGE_LIST = "list_images"
        cmdnames.LANDMARK_PREV = "previous_landmark"
        cmdnames.LANDMARK_NEXT = "next_landmark"
        cmdnames.LANDMARK_LIST = "list_landmarks"
        cmdnames.LIST_PREV = "previous_list"
        cmdnames.LIST_NEXT = "next_list"
        cmdnames.LIST_LIST = "list_lists"
        cmdnames.LIST_ITEM_PREV = "previous_list_item"
        cmdnames.LIST_ITEM_NEXT = "next_list_item"
        cmdnames.LIST_ITEM_LIST = "list_list_items"
        cmdnames.LIVE_REGION_PREV = "previous_live_region"
        cmdnames.LIVE_REGION_NEXT = "next_live_region"
        cmdnames.LIVE_REGION_LAST = "last_live_region"
        cmdnames.PARAGRAPH_PREV = "previous_paragraph"
        cmdnames.PARAGRAPH_NEXT = "next_paragraph"
        cmdnames.PARAGRAPH_LIST = "list_paragraphs"
        cmdnames.RADIO_BUTTON_PREV = "previous_radio_button"
        cmdnames.RADIO_BUTTON_NEXT = "next_radio_button"
        cmdnames.RADIO_BUTTON_LIST = "list_radio_buttons"
        cmdnames.SEPARATOR_PREV = "previous_separator"
        cmdnames.SEPARATOR_NEXT = "next_separator"
        cmdnames.TABLE_PREV = "previous_table"
        cmdnames.TABLE_NEXT = "next_table"
        cmdnames.TABLE_LIST = "list_tables"
        cmdnames.UNVISITED_LINK_PREV = "previous_unvisited_link"
        cmdnames.UNVISITED_LINK_NEXT = "next_unvisited_link"
        cmdnames.UNVISITED_LINK_LIST = "list_unvisited_links"
        cmdnames.VISITED_LINK_PREV = "previous_visited_link"
        cmdnames.VISITED_LINK_NEXT = "next_visited_link"
        cmdnames.VISITED_LINK_LIST = "list_visited_links"
        cmdnames.LINK_PREV = "previous_link"
        cmdnames.LINK_NEXT = "next_link"
        cmdnames.LINK_LIST = "list_links"
        cmdnames.CLICKABLE_PREV = "previous_clickable"
        cmdnames.CLICKABLE_NEXT = "next_clickable"
        cmdnames.CLICKABLE_LIST = "list_clickables"
        cmdnames.LARGE_OBJECT_PREV = "previous_large_object"
        cmdnames.LARGE_OBJECT_NEXT = "next_large_object"
        cmdnames.LARGE_OBJECT_LIST = "list_large_objects"
        cmdnames.CONTAINER_START = "container_start"
        cmdnames.CONTAINER_END = "container_end"

        essential_modules["orca.orca_i18n"]._ = lambda x: x
        essential_modules["orca.debug"].print_message = test_context.Mock()
        essential_modules["orca.debug"].LEVEL_INFO = 800

        controller_mock = test_context.Mock()
        controller_mock.register_decorated_module.return_value = None
        essential_modules["orca.dbus_service"].get_remote_controller.return_value = controller_mock

        focus_manager_instance = test_context.Mock()
        focus_manager_instance.get_locus_of_focus.return_value = None
        essential_modules["orca.focus_manager"].get_manager.return_value = focus_manager_instance

        essential_modules["orca.AXObject"].supports_collection.return_value = True
        essential_modules["orca.AXUtilities"].is_heading.return_value = False

        return essential_modules

    @pytest.mark.parametrize(
        "direction,event_provided,context_available,expected_result",
        [
            pytest.param("next", False, True, True, id="next_char_no_event_returns_true"),
            pytest.param("next", True, False, False, id="next_char_no_context_returns_false"),
            pytest.param("next", True, True, True, id="next_char_valid_navigation_succeeds"),
            pytest.param("previous", False, True, True, id="prev_char_no_event_returns_true"),
            pytest.param("previous", True, False, False, id="prev_char_no_context_returns_false"),
            pytest.param("previous", True, True, True, id="prev_char_valid_navigation_succeeds"),
        ],
    )
    def test_character_navigation(
        self,
        test_context: OrcaTestContext,
        direction: str,
        event_provided: bool,
        context_available: bool,
        expected_result: bool,
    ) -> None:
        """Test character navigation (next/previous) with various conditions."""

        essential_modules = self._setup_dependencies(test_context)
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        ax_object_mock = essential_modules["orca.ax_object"]
        ax_object_mock.AXObject.supports_text.side_effect = lambda obj: obj is not None
        ax_object_mock.AXObject.is_valid.side_effect = lambda obj: obj is not None
        ax_object_mock.AXObject.is_ancestor.side_effect = (
            lambda obj, root, same: obj is not None and root is not None
        )

        navigator = CaretNavigator()
        test_context.patch_object(navigator, "_get_root_object", return_value=None)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock() if event_provided else None

        if context_available:
            mock_obj = test_context.Mock()
            if direction == "next":
                mock_script.utilities.next_context.return_value = (mock_obj, 10)
            else:
                mock_script.utilities.previous_context.return_value = (mock_obj, 5)
        elif direction == "next":
            mock_script.utilities.next_context.return_value = (None, 0)
        else:
            mock_script.utilities.previous_context.return_value = (None, 0)

        navigation_method = getattr(navigator, f"{direction}_character")
        result = navigation_method(mock_script, mock_event)
        assert result == expected_result

        if expected_result:
            pres_manager = essential_modules["orca.presentation_manager"].get_manager()
            assert navigator._last_input_event == mock_event
            mock_script.utilities.set_caret_position.assert_called_once()
            pres_manager.interrupt_presentation.assert_called_once()
            mock_script.update_braille.assert_called_once()
            mock_script.say_character.assert_called_once()

    @pytest.mark.parametrize(
        "direction,context_result,word_contents,expected_result",
        [
            pytest.param("next", (None, 0), None, False, id="next_word_no_context"),
            pytest.param("next", ("obj", 20), [], False, id="next_word_no_contents"),
            pytest.param(
                "next",
                ("obj", 20),
                [("obj", 20, 25, "word")],
                True,
                id="next_word_success",
            ),
            pytest.param("previous", (None, 0), None, False, id="previous_word_no_context"),
            pytest.param("previous", ("obj", 15), [], False, id="previous_word_no_contents"),
            pytest.param(
                "previous",
                ("obj", 15),
                [("obj", 10, 15, "word")],
                True,
                id="previous_word_success",
            ),
        ],
    )
    def test_word_navigation(
        self,
        test_context: OrcaTestContext,
        direction: str,
        context_result: tuple,
        word_contents: list | None,
        expected_result: bool,
    ) -> None:
        """Test word navigation (next/previous) with various error conditions."""

        essential_modules = self._setup_dependencies(test_context)
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()

        if direction == "next":
            mock_script.utilities.next_context.return_value = context_result
        else:
            mock_script.utilities.previous_context.return_value = context_result

        mock_script.utilities.get_word_contents_at_offset.return_value = word_contents or []

        mock_script.utilities.set_caret_position = test_context.Mock()
        mock_script.update_braille = test_context.Mock()
        mock_script.say_word = test_context.Mock()

        navigation_method = getattr(navigator, f"{direction}_word")
        result = navigation_method(mock_script, mock_event)

        assert result == expected_result

        pres_manager = essential_modules["orca.presentation_manager"].get_manager()
        if expected_result:
            assert navigator._last_input_event == mock_event
            mock_script.utilities.set_caret_position.assert_called()
            pres_manager.interrupt_presentation.assert_called_once()
            mock_script.update_braille.assert_called_once()
            mock_script.say_word.assert_called_once()
        else:
            mock_script.utilities.set_caret_position.assert_not_called()
            pres_manager.interrupt_presentation.assert_not_called()

    @pytest.mark.parametrize(
        "test_method,expected_result",
        [
            pytest.param("suspend_commands", True, id="suspend_commands"),
            pytest.param("toggle_enabled", True, id="toggle_enabled"),
        ],
    )
    def test_navigator_control_methods(
        self,
        test_context: OrcaTestContext,
        test_method: str,
        expected_result: bool,
    ) -> None:
        """Test CaretNavigator control methods."""

        essential_modules = self._setup_dependencies(test_context)
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()

        if test_method == "suspend_commands":
            mock_cmd_mgr = test_context.Mock()
            essential_modules["orca.command_manager"].get_manager.return_value = mock_cmd_mgr
            test_context.patch_object(navigator, "_is_active_script", return_value=True)
            navigator._suspended = False
            navigator.suspend_commands(mock_script, True, "test reason")
            assert navigator._suspended == expected_result
            mock_cmd_mgr.set_group_suspended.assert_called_once()

        elif test_method == "toggle_enabled":
            mock_cmd_mgr = test_context.Mock()
            essential_modules["orca.command_manager"].get_manager.return_value = mock_cmd_mgr

            guilabels_mock = essential_modules["orca.guilabels"]
            guilabels_mock.CARET_NAVIGATION_ENABLED = "Caret navigation enabled"
            guilabels_mock.CARET_NAVIGATION_DISABLED = "Caret navigation disabled"

            result = navigator.toggle_enabled(mock_script, mock_event)
            assert result == expected_result
            mock_cmd_mgr.set_group_enabled.assert_called_once()

    def test_navigator_initialization(self, test_context: OrcaTestContext) -> None:
        """Test CaretNavigator initialization."""
        self._setup_dependencies(test_context)
        from orca import command_manager
        from orca.caret_navigator import CaretNavigator

        navigator = CaretNavigator()

        assert hasattr(navigator, "_last_input_event")
        assert hasattr(navigator, "_suspended")
        assert navigator._last_input_event is None
        assert navigator._suspended is False
        # Commands are registered in CommandManager
        cmd_manager = command_manager.get_manager()
        assert cmd_manager is not None

    @pytest.mark.parametrize(
        "navigation_type,in_say_all,current_line,next_prev_contents,expected_result",
        [
            pytest.param(
                "next_line",
                True,
                [("obj", 0, 10, "text")],
                [],
                True,
                id="next_line_in_say_all",
            ),
            pytest.param("next_line", False, [], [], False, id="next_line_no_current_line"),
            pytest.param(
                "next_line",
                False,
                [("obj", 0, 10, "text")],
                [],
                False,
                id="next_line_no_next_contents",
            ),
            pytest.param(
                "next_line",
                False,
                [("obj", 0, 10, "text")],
                [("obj2", 11, 21, "next")],
                True,
                id="next_line_success",
            ),
            pytest.param(
                "previous_line",
                True,
                [("obj", 0, 10, "text")],
                [],
                True,
                id="previous_line_in_say_all",
            ),
            pytest.param("previous_line", False, [], [], False, id="previous_line_no_contents"),
            pytest.param(
                "previous_line",
                False,
                [("obj", 0, 10, "text")],
                [("obj", 0, 10, "prev")],
                True,
                id="previous_line_success",
            ),
        ],
    )
    def test_line_navigation(
        self,
        test_context: OrcaTestContext,
        navigation_type: str,
        in_say_all: bool,
        current_line: list | None,
        next_prev_contents: list,
        expected_result: bool,
    ) -> None:
        """Test line navigation including say-all mode handling."""

        essential_modules = self._setup_dependencies(test_context)
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()

        focus_manager_mock = essential_modules["orca.focus_manager"]
        manager_instance = test_context.Mock()
        focus_manager_mock.get_manager.return_value = manager_instance
        manager_instance.in_say_all.return_value = in_say_all

        if in_say_all:
            from orca import say_all_presenter  # pylint: disable=import-outside-toplevel

            say_all_presenter.get_presenter().set_rewind_and_fast_forward_enabled(True)

        if navigation_type == "next_line" and not in_say_all:
            mock_script.utilities.get_caret_context.return_value = ("obj", 5)
            mock_script.utilities.get_line_contents_at_offset.return_value = current_line
            mock_script.utilities.get_next_line_contents.return_value = next_prev_contents
            test_context.patch_object(navigator, "_get_end_of_file", return_value=(None, -1))
            test_context.patch_object(navigator, "_line_contains_context", return_value=False)
            test_context.patch_object(navigator, "_is_navigable_object", return_value=True)
        elif navigation_type == "previous_line" and not in_say_all:
            mock_script.utilities.get_caret_context.return_value = ("obj", 5)
            mock_script.utilities.get_line_contents_at_offset.return_value = current_line
            mock_script.utilities.get_previous_line_contents.return_value = next_prev_contents
            test_context.patch_object(navigator, "_get_start_of_file", return_value=(None, -1))
            test_context.patch_object(navigator, "_line_contains_context", return_value=False)
            test_context.patch_object(navigator, "_is_navigable_object", return_value=True)

        mock_script.utilities.set_caret_position = test_context.Mock()

        pres_manager = essential_modules["orca.presentation_manager"].get_manager()
        pres_manager.interrupt_presentation.reset_mock()
        pres_manager.speak_contents.reset_mock()
        pres_manager.display_contents.reset_mock()

        navigation_method = getattr(navigator, f"{navigation_type}")
        result = navigation_method(mock_script, mock_event)

        assert result == expected_result

        if expected_result and not in_say_all:
            assert navigator._last_input_event == mock_event
            mock_script.utilities.set_caret_position.assert_called()
            pres_manager.interrupt_presentation.assert_called_once()
            pres_manager.speak_contents.assert_called_once()
            pres_manager.display_contents.assert_called_once()
        elif in_say_all:
            assert navigator._last_input_event != mock_event

    @pytest.mark.parametrize(
        "navigation_type,line_contents,expected_result",
        [
            pytest.param("start_of_line", [], False, id="start_of_line_no_line"),
            pytest.param(
                "start_of_line",
                [("obj", 5, 15, "text")],
                True,
                id="start_of_line_success",
            ),
            pytest.param("end_of_line", [], False, id="end_of_line_no_line"),
            pytest.param(
                "end_of_line",
                [("obj", 5, 15, "text ")],
                True,
                id="end_of_line_with_space",
            ),
            pytest.param("end_of_line", [("obj", 5, 15, "text")], True, id="end_of_line_no_space"),
        ],
    )
    def test_line_boundary_navigation(
        self,
        test_context: OrcaTestContext,
        navigation_type: str,
        line_contents: list,
        expected_result: bool,
    ) -> None:
        """Test start/end of line navigation."""

        essential_modules = self._setup_dependencies(test_context)
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()

        mock_script.utilities.get_caret_context.return_value = ("obj", 10)
        mock_script.utilities.get_line_contents_at_offset.return_value = line_contents

        mock_script.utilities.set_caret_position = test_context.Mock()
        mock_script.say_character = test_context.Mock()

        pres_manager = essential_modules["orca.presentation_manager"].get_manager()
        pres_manager.interrupt_presentation.reset_mock()
        pres_manager.display_contents.reset_mock()

        navigation_method = getattr(navigator, f"{navigation_type}")
        result = navigation_method(mock_script, mock_event)

        assert result == expected_result

        if expected_result:
            assert navigator._last_input_event == mock_event
            mock_script.utilities.set_caret_position.assert_called()
            pres_manager.interrupt_presentation.assert_called_once()
            mock_script.say_character.assert_called_once()
            pres_manager.display_contents.assert_called_once()

    @pytest.mark.parametrize(
        "script_is_active,expected_result",
        [
            pytest.param(True, True, id="script_is_active"),
            pytest.param(False, False, id="script_is_not_active"),
        ],
    )
    def test_is_active_script(
        self,
        test_context: OrcaTestContext,
        script_is_active: bool,
        expected_result: bool,
    ) -> None:
        """Test _is_active_script method with active and non-active scripts."""

        essential_modules = self._setup_dependencies(test_context)
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        mock_script = test_context.Mock()
        mock_active_script = test_context.Mock()

        script_manager_mock = essential_modules["orca.script_manager"]
        manager_instance = test_context.Mock()
        script_manager_mock.get_manager.return_value = manager_instance

        if script_is_active:
            manager_instance.get_active_script.return_value = mock_script
        else:
            manager_instance.get_active_script.return_value = mock_active_script

        debug_mock = essential_modules["orca.debug"]
        debug_mock.LEVEL_INFO = 800
        debug_mock.print_tokens = test_context.Mock()

        result = navigator._is_active_script(mock_script)
        assert result == expected_result

        if not script_is_active:
            debug_mock.print_tokens.assert_called_once()

    def test_get_is_enabled(self, test_context: OrcaTestContext) -> None:
        """Test get_is_enabled returns setting value."""

        self._setup_dependencies(test_context)
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        result = navigator.get_is_enabled()
        assert result is True

    def test_set_is_enabled_no_change(self, test_context: OrcaTestContext) -> None:
        """Test set_is_enabled still calls set_group_enabled even if value unchanged."""

        essential_modules = self._setup_dependencies(test_context)
        mock_cmd_mgr = test_context.Mock()
        essential_modules["orca.command_manager"].get_manager.return_value = mock_cmd_mgr
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        result = navigator.set_is_enabled(True)
        assert result is True
        mock_cmd_mgr.set_group_enabled.assert_called_once()

    def test_set_is_enabled_updates_setting(self, test_context: OrcaTestContext) -> None:
        """Test set_is_enabled updates setting and calls CommandManager."""

        essential_modules = self._setup_dependencies(test_context)
        from orca import gsettings_registry

        gsettings_registry.get_registry().set_runtime_value("caret-navigation", "enabled", False)
        mock_script = test_context.Mock()
        essential_modules[
            "orca.script_manager"
        ].get_manager.return_value.get_active_script.return_value = mock_script
        mock_cmd_mgr = test_context.Mock()
        essential_modules["orca.command_manager"].get_manager.return_value = mock_cmd_mgr
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()

        result = navigator.set_is_enabled(True)
        assert result is True
        assert navigator.get_is_enabled() is True
        assert navigator._last_input_event is None
        mock_cmd_mgr.set_group_enabled.assert_called_once()

    def test_set_is_enabled_no_active_script(self, test_context: OrcaTestContext) -> None:
        """Test set_is_enabled updates state even with no active script."""

        essential_modules = self._setup_dependencies(test_context)
        from orca import gsettings_registry

        gsettings_registry.get_registry().set_runtime_value("caret-navigation", "enabled", False)
        essential_modules[
            "orca.script_manager"
        ].get_manager.return_value.get_active_script.return_value = None
        mock_cmd_mgr = test_context.Mock()
        essential_modules["orca.command_manager"].get_manager.return_value = mock_cmd_mgr
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()

        result = navigator.set_is_enabled(True)
        assert result is True
        mock_cmd_mgr.set_group_enabled.assert_called_once()

    def test_get_triggers_focus_mode(self, test_context: OrcaTestContext) -> None:
        """Test get_triggers_focus_mode returns setting value."""

        self._setup_dependencies(test_context)
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        result = navigator.get_triggers_focus_mode()
        assert result is False

    def test_set_triggers_focus_mode(self, test_context: OrcaTestContext) -> None:
        """Test set_triggers_focus_mode updates setting."""

        self._setup_dependencies(test_context)
        from orca import gsettings_registry

        gsettings_registry.get_registry().set_runtime_value(
            "caret-navigation",
            "triggers-focus-mode",
            True,
        )
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        result = navigator.set_triggers_focus_mode(False)
        assert result is True
        assert navigator.get_triggers_focus_mode() is False

    def test_set_triggers_focus_mode_no_change(self, test_context: OrcaTestContext) -> None:
        """Test set_triggers_focus_mode returns early if unchanged."""

        self._setup_dependencies(test_context)
        from orca import gsettings_registry

        gsettings_registry.get_registry().set_runtime_value(
            "caret-navigation",
            "triggers-focus-mode",
            True,
        )
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        result = navigator.set_triggers_focus_mode(True)
        assert result is True
        # set_setting no longer used - settings are set directly

    def test_get_enabled_for_script(self, test_context: OrcaTestContext) -> None:
        """Test get_enabled_for_script returns script-specific state."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        navigator._enabled_for_script[mock_script] = True
        result = navigator.get_enabled_for_script(mock_script)
        assert result is True

    def test_get_enabled_for_script_default(self, test_context: OrcaTestContext) -> None:
        """Test get_enabled_for_script returns False by default."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        result = navigator.get_enabled_for_script(mock_script)
        assert result is False

    def test_set_enabled_for_script(self, test_context: OrcaTestContext) -> None:
        """Test set_enabled_for_script updates script-specific state and calls set_is_enabled."""

        essential_modules = self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        from orca import gsettings_registry

        gsettings_registry.get_registry().set_runtime_value("caret-navigation", "enabled", False)
        essential_modules[
            "orca.script_manager"
        ].get_manager.return_value.get_active_script.return_value = mock_script
        mock_cmd_mgr = test_context.Mock()
        essential_modules["orca.command_manager"].get_manager.return_value = mock_cmd_mgr
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        test_context.patch_object(navigator, "_is_active_script", return_value=True)

        navigator.set_enabled_for_script(mock_script, True)
        assert navigator._enabled_for_script[mock_script] is True
        mock_cmd_mgr.set_group_enabled.assert_called_once()

    def test_set_enabled_for_script_inactive_script(self, test_context: OrcaTestContext) -> None:
        """Test set_enabled_for_script doesn't call set_group_enabled for inactive script."""

        essential_modules = self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_cmd_mgr = test_context.Mock()
        essential_modules["orca.command_manager"].get_manager.return_value = mock_cmd_mgr
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        test_context.patch_object(navigator, "_is_active_script", return_value=False)

        navigator.set_enabled_for_script(mock_script, True)
        assert navigator._enabled_for_script[mock_script] is True
        mock_cmd_mgr.set_group_enabled.assert_not_called()

    def test_set_enabled_for_script_always_calls_set_group_enabled(
        self,
        test_context: OrcaTestContext,
    ) -> None:
        """Test set_enabled_for_script always calls set_group_enabled even if setting matches.

        This is a regression test for issue #655 where caret navigation commands
        were not being enabled because set_is_enabled() would early-return when
        the setting already matched the desired value.
        """

        essential_modules = self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        essential_modules[
            "orca.script_manager"
        ].get_manager.return_value.get_active_script.return_value = mock_script
        mock_cmd_mgr = test_context.Mock()
        essential_modules["orca.command_manager"].get_manager.return_value = mock_cmd_mgr
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        test_context.patch_object(navigator, "_is_active_script", return_value=True)

        navigator.set_enabled_for_script(mock_script, True)
        assert navigator._enabled_for_script[mock_script] is True
        mock_cmd_mgr.set_group_enabled.assert_called_once()

    def test_last_command_prevents_focus_mode_true(self, test_context: OrcaTestContext) -> None:
        """Test last_command_prevents_focus_mode returns True."""

        self._setup_dependencies(test_context)
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        mock_event = test_context.Mock()
        navigator._last_input_event = mock_event
        test_context.patch_object(
            navigator,
            "last_input_event_was_navigation_command",
            return_value=True,
        )
        result = navigator.last_command_prevents_focus_mode()
        assert result is True

    def test_last_command_prevents_focus_mode_false_no_event(
        self,
        test_context: OrcaTestContext,
    ) -> None:
        """Test last_command_prevents_focus_mode returns False if no event."""

        self._setup_dependencies(test_context)
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        navigator._last_input_event = None
        result = navigator.last_command_prevents_focus_mode()
        assert result is False

    def test_last_command_prevents_focus_mode_false_setting_true(
        self,
        test_context: OrcaTestContext,
    ) -> None:
        """Test last_command_prevents_focus_mode returns False if setting True."""

        self._setup_dependencies(test_context)
        from orca.caret_navigator import CaretNavigator  # pylint: disable=import-outside-toplevel

        navigator = CaretNavigator()
        navigator.set_triggers_focus_mode(True)
        mock_event = test_context.Mock()
        navigator._last_input_event = mock_event
        test_context.patch_object(
            navigator,
            "last_input_event_was_navigation_command",
            return_value=True,
        )
        result = navigator.last_command_prevents_focus_mode()
        assert result is False

    def test_successful_navigation_emits_region_changed(
        self,
        test_context: OrcaTestContext,
    ) -> None:
        """Test successful caret navigation emits region_changed with CARET_NAVIGATOR mode."""

        essential_modules = self._setup_dependencies(test_context)
        from orca import focus_manager
        from orca.caret_navigator import CaretNavigator

        ax_object_mock = essential_modules["orca.ax_object"]
        ax_object_mock.AXObject.supports_text.side_effect = lambda obj: obj is not None
        ax_object_mock.AXObject.is_valid.side_effect = lambda obj: obj is not None
        ax_object_mock.AXObject.is_ancestor.side_effect = (
            lambda obj, root, same: obj is not None and root is not None
        )

        focus_manager_mock = essential_modules["orca.focus_manager"]
        manager_instance = test_context.Mock()
        focus_manager_mock.get_manager.return_value = manager_instance
        focus_manager_mock.CARET_NAVIGATOR = focus_manager.CARET_NAVIGATOR

        navigator = CaretNavigator()
        test_context.patch_object(navigator, "_get_root_object", return_value=None)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        mock_obj = test_context.Mock()

        mock_script.utilities.next_context.return_value = (mock_obj, 10)

        result = navigator.next_character(mock_script, mock_event)

        assert result is True
        manager_instance.emit_region_changed.assert_called()
        call_kwargs = manager_instance.emit_region_changed.call_args
        assert call_kwargs.kwargs.get("mode") == focus_manager.CARET_NAVIGATOR



# Unit tests for sound_presenter.py methods.
#
# Copyright 2025 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=wrong-import-position
# pylint: disable=import-outside-toplevel
# pylint: disable=protected-access

"""Unit tests for sound_presenter.py methods."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from unittest.mock import MagicMock

    from .orca_test_context import OrcaTestContext


@pytest.mark.unit
class TestSoundPresenter:
    """Test SoundPresenter class methods."""

    def _setup_dependencies(self, test_context: OrcaTestContext) -> dict[str, MagicMock]:
        """Set up mocks for sound_presenter dependencies."""

        essential_modules = test_context.setup_shared_dependencies([])

        debug_mock = essential_modules["orca.debug"]
        debug_mock.print_message = test_context.Mock()
        debug_mock.LEVEL_INFO = 800

        dbus_service_mock = essential_modules["orca.dbus_service"]
        controller_mock = test_context.Mock()
        controller_mock.register_decorated_module = test_context.Mock()
        dbus_service_mock.get_remote_controller = test_context.Mock(return_value=controller_mock)
        dbus_service_mock.getter = lambda func: func
        dbus_service_mock.setter = lambda func: func

        essential_modules["controller"] = controller_mock

        from orca import gsettings_registry

        registry = gsettings_registry.get_registry()
        registry.clear_runtime_values()

        return essential_modules

    def test_initialization_registers_dbus_module(self, test_context: OrcaTestContext) -> None:
        """Test that SoundPresenter registers itself with D-Bus on initialization."""

        mocks = self._setup_dependencies(test_context)

        from orca import sound_presenter

        presenter = sound_presenter.SoundPresenter()
        mocks["controller"].register_decorated_module.assert_called_with(
            "SoundPresenter",
            presenter,
        )

    def test_get_sound_is_enabled_returns_setting(self, test_context: OrcaTestContext) -> None:
        """Test get_sound_is_enabled returns the enableSound setting."""

        self._setup_dependencies(test_context)

        from orca.sound_presenter import SoundPresenter

        presenter = SoundPresenter()
        assert presenter.get_sound_is_enabled() is True

        presenter.set_sound_is_enabled(False)
        assert presenter.get_sound_is_enabled() is False

    def test_set_sound_is_enabled_updates_setting(self, test_context: OrcaTestContext) -> None:
        """Test set_sound_is_enabled updates the enableSound setting."""

        self._setup_dependencies(test_context)

        from orca.sound_presenter import SoundPresenter

        presenter = SoundPresenter()
        result = presenter.set_sound_is_enabled(False)

        assert result is True
        assert presenter.get_sound_is_enabled() is False

    def test_get_sound_volume_returns_setting(self, test_context: OrcaTestContext) -> None:
        """Test get_sound_volume returns the soundVolume setting."""

        self._setup_dependencies(test_context)

        from orca.sound_presenter import SoundPresenter

        presenter = SoundPresenter()
        presenter.set_sound_volume(0.75)
        assert presenter.get_sound_volume() == 0.75

    def test_set_sound_volume_updates_setting(self, test_context: OrcaTestContext) -> None:
        """Test set_sound_volume updates the soundVolume setting."""

        self._setup_dependencies(test_context)

        from orca.sound_presenter import SoundPresenter

        presenter = SoundPresenter()
        result = presenter.set_sound_volume(0.8)

        assert result is True
        assert presenter.get_sound_volume() == 0.8

    def test_get_beep_progress_bar_updates_returns_setting(
        self,
        test_context: OrcaTestContext,
    ) -> None:
        """Test get_beep_progress_bar_updates returns the beepProgressBarUpdates setting."""

        self._setup_dependencies(test_context)

        from orca.sound_presenter import SoundPresenter

        presenter = SoundPresenter()
        presenter.set_beep_progress_bar_updates(True)
        assert presenter.get_beep_progress_bar_updates() is True

        presenter.set_beep_progress_bar_updates(False)
        assert presenter.get_beep_progress_bar_updates() is False

    def test_set_beep_progress_bar_updates_updates_setting(
        self,
        test_context: OrcaTestContext,
    ) -> None:
        """Test set_beep_progress_bar_updates updates the beepProgressBarUpdates setting."""

        self._setup_dependencies(test_context)

        from orca.sound_presenter import SoundPresenter

        presenter = SoundPresenter()
        result = presenter.set_beep_progress_bar_updates(True)

        assert result is True
        assert presenter.get_beep_progress_bar_updates() is True

    def test_get_progress_bar_beep_interval_returns_setting(
        self,
        test_context: OrcaTestContext,
    ) -> None:
        """Test get_progress_bar_beep_interval returns the progressBarBeepInterval setting."""

        self._setup_dependencies(test_context)

        from orca.sound_presenter import SoundPresenter

        presenter = SoundPresenter()
        presenter.set_progress_bar_beep_interval(10)
        assert presenter.get_progress_bar_beep_interval() == 10

    def test_set_progress_bar_beep_interval_updates_setting(
        self,
        test_context: OrcaTestContext,
    ) -> None:
        """Test set_progress_bar_beep_interval updates the progressBarBeepInterval setting."""

        self._setup_dependencies(test_context)

        from orca.sound_presenter import SoundPresenter

        presenter = SoundPresenter()
        result = presenter.set_progress_bar_beep_interval(15)

        assert result is True
        assert presenter.get_progress_bar_beep_interval() == 15

    def test_get_progress_bar_beep_verbosity_returns_setting(
        self,
        test_context: OrcaTestContext,
    ) -> None:
        """Test get_progress_bar_beep_verbosity returns the progressBarBeepVerbosity setting."""

        self._setup_dependencies(test_context)

        from orca.sound_presenter import SoundPresenter

        presenter = SoundPresenter()
        presenter.set_progress_bar_beep_verbosity(2)
        assert presenter.get_progress_bar_beep_verbosity() == 2

    def test_set_progress_bar_beep_verbosity_updates_setting(
        self,
        test_context: OrcaTestContext,
    ) -> None:
        """Test set_progress_bar_beep_verbosity updates the progressBarBeepVerbosity setting."""

        self._setup_dependencies(test_context)

        from orca.sound_presenter import SoundPresenter

        presenter = SoundPresenter()
        result = presenter.set_progress_bar_beep_verbosity(2)

        assert result is True
        assert presenter.get_progress_bar_beep_verbosity() == 2

    def test_get_presenter_returns_singleton(self, test_context: OrcaTestContext) -> None:
        """Test get_presenter returns the module-level presenter instance."""

        self._setup_dependencies(test_context)

        from orca import sound_presenter

        presenter1 = sound_presenter.get_presenter()
        presenter2 = sound_presenter.get_presenter()

        assert presenter1 is presenter2
        assert isinstance(presenter1, sound_presenter.SoundPresenter)



# Unit tests for preferences_grid_base.py.
#
# Copyright 2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=wrong-import-position
# pylint: disable=protected-access
# pylint: disable=import-outside-toplevel

"""Unit tests for preferences_grid_base.py."""

from __future__ import annotations

from typing import Any

import gi
import pytest

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

from orca.preferences_grid_base import (
    AutoPreferencesGrid,
    BooleanPreferenceControl,
    CategoryListBoxRow,
    CommandListBoxRow,
    ControlType,
    EnumPreferenceControl,
    FloatRangePreferenceControl,
    FocusManagedListBox,
    IntRangePreferenceControl,
    PreferencesGridBase,
    RadioButtonWithActions,
    SelectionPreferenceControl,
    StackedPreferencesHelper,
)


@pytest.mark.unit
class TestPreferenceControlDataclasses:
    """Test preference control dataclass creation and properties."""

    def test_boolean_preference_control_creation(self) -> None:
        """Test BooleanPreferenceControl can be created with required fields."""

        getter_called: list[bool] = [False]
        setter_value: list[bool | None] = [None]

        def getter() -> bool:
            getter_called[0] = True
            return True

        def setter(value: bool) -> bool:
            setter_value[0] = value
            return True

        control = BooleanPreferenceControl(
            label="Test Label",
            getter=getter,
            setter=setter,
            prefs_key="testKey",
        )

        assert control.label == "Test Label"
        assert control.prefs_key == "testKey"
        assert control.getter() is True
        assert getter_called[0] is True
        control.setter(False)
        assert setter_value[0] is False

    def test_boolean_preference_control_optional_fields(self) -> None:
        """Test BooleanPreferenceControl optional fields have correct defaults."""

        control = BooleanPreferenceControl(label="Test", getter=lambda: True, setter=lambda x: True)

        assert control.prefs_key is None
        assert control.member_of is None
        assert control.determine_sensitivity is None
        assert control.apply_immediately is True

    def test_int_range_preference_control_creation(self) -> None:
        """Test IntRangePreferenceControl can be created with all fields."""

        control = IntRangePreferenceControl(
            label="Volume",
            minimum=0,
            maximum=100,
            getter=lambda: 50,
            setter=lambda x: True,
            prefs_key="volume",
        )

        assert control.label == "Volume"
        assert control.minimum == 0
        assert control.maximum == 100
        assert control.getter() == 50

    def test_float_range_preference_control_creation(self) -> None:
        """Test FloatRangePreferenceControl can be created with all fields."""

        control = FloatRangePreferenceControl(
            label="Rate",
            minimum=0.0,
            maximum=1.0,
            getter=lambda: 0.5,
            setter=lambda x: True,
        )

        assert control.label == "Rate"
        assert control.minimum == 0.0
        assert control.maximum == 1.0
        assert control.getter() == 0.5

    def test_enum_preference_control_creation(self) -> None:
        """Test EnumPreferenceControl can be created with options and values."""

        control = EnumPreferenceControl(
            label="Style",
            options=["Option A", "Option B", "Option C"],
            values=[1, 2, 3],
            getter=lambda: 2,
            setter=lambda x: True,
        )

        assert control.label == "Style"
        assert len(control.options) == 3
        assert control.values == [1, 2, 3]
        assert control.getter() == 2

    def test_selection_preference_control_creation(self) -> None:
        """Test SelectionPreferenceControl can be created with actions callback."""

        def get_actions(_value: Any) -> list[tuple[str, str, Any]]:
            return [("Edit", "edit-symbolic", lambda: None)]

        control = SelectionPreferenceControl(
            label="Profile",
            options=["Default", "Custom"],
            getter=lambda: "Default",
            setter=lambda x: True,
            get_actions_for_option=get_actions,
        )

        assert control.label == "Profile"
        assert len(control.options) == 2
        assert control.get_actions_for_option is not None
        actions = control.get_actions_for_option("Default")
        assert len(actions) == 1
        assert actions[0][0] == "Edit"


@pytest.mark.unit
class TestHelperClasses:
    """Test helper classes in preferences_grid_base."""

    def test_category_list_box_row_stores_category(self) -> None:
        """Test CategoryListBoxRow stores and returns category identifier."""

        row = CategoryListBoxRow("speech_settings")
        assert row.category == "speech_settings"

    def test_category_list_box_row_is_gtk_list_box_row(self) -> None:
        """Test CategoryListBoxRow inherits from Gtk.ListBoxRow."""

        row = CategoryListBoxRow("test")
        assert isinstance(row, Gtk.ListBoxRow)

    def test_command_list_box_row_properties(self) -> None:
        """Test CommandListBoxRow property getters and setters."""

        row = CommandListBoxRow()
        assert row.command is None
        assert row.vbox is None
        assert row.binding_label is None

        mock_command = object()
        mock_vbox = Gtk.Box()
        mock_label = Gtk.Label()

        row.command = mock_command
        row.vbox = mock_vbox
        row.binding_label = mock_label

        assert row.command is mock_command
        assert row.vbox is mock_vbox
        assert row.binding_label is mock_label

    def test_radio_button_with_actions_stores_buttons(self) -> None:
        """Test RadioButtonWithActions stores action buttons list."""

        radio = RadioButtonWithActions(label="Test Option")
        assert not radio.action_buttons

        button1 = Gtk.Button()
        button2 = Gtk.Button()
        radio.action_buttons = [button1, button2]

        assert len(radio.action_buttons) == 2
        assert radio.action_buttons[0] is button1

    def test_radio_button_with_actions_is_gtk_radio_button(self) -> None:
        """Test RadioButtonWithActions inherits from Gtk.RadioButton."""

        radio = RadioButtonWithActions(label="Test")
        assert isinstance(radio, Gtk.RadioButton)

    def test_stacked_preferences_helper_initial_state(self) -> None:
        """Test StackedPreferencesHelper initializes with None values."""

        helper = StackedPreferencesHelper()
        assert helper.stack is None
        assert helper.categories_listbox is None
        assert helper.detail_listbox is None
        assert not helper.disable_widgets
        assert helper.on_category_activated_callback is None

    def test_stacked_preferences_helper_register_widgets(self) -> None:
        """Test StackedPreferencesHelper can register widgets to disable."""

        helper = StackedPreferencesHelper()
        button1 = Gtk.Button()
        button2 = Gtk.Button()

        helper.register_disable_widgets(button1, button2)
        assert len(helper.disable_widgets) == 2
        assert button1 in helper.disable_widgets

    def test_stacked_preferences_helper_show_categories(self) -> None:
        """Test StackedPreferencesHelper show_categories enables registered widgets."""

        helper = StackedPreferencesHelper()
        helper.stack = Gtk.Stack()
        cat_label = Gtk.Label(label="Categories")
        det_label = Gtk.Label(label="Detail")
        helper.stack.add_named(cat_label, "categories")
        helper.stack.add_named(det_label, "detail")

        # Show the stack to ensure children are realized
        helper.stack.show_all()  # pylint: disable=no-member
        helper.stack.set_visible_child_name("detail")

        button = Gtk.Button()
        button.set_sensitive(False)
        helper.register_disable_widgets(button)

        helper.show_categories()
        assert button.get_sensitive() is True
        # Verify the visible child is the categories child
        assert helper.stack.get_visible_child() is cat_label

    def test_stacked_preferences_helper_show_detail(self) -> None:
        """Test StackedPreferencesHelper show_detail disables registered widgets."""

        helper = StackedPreferencesHelper()
        helper.stack = Gtk.Stack()
        cat_label = Gtk.Label(label="Categories")
        det_label = Gtk.Label(label="Detail")
        helper.stack.add_named(cat_label, "categories")
        helper.stack.add_named(det_label, "detail")

        # Show the stack to ensure children are realized
        helper.stack.show_all()  # pylint: disable=no-member
        helper.stack.set_visible_child_name("categories")

        button = Gtk.Button()
        button.set_sensitive(True)
        helper.register_disable_widgets(button)

        helper.show_detail()
        assert button.get_sensitive() is False
        # Verify the visible child is the detail child
        assert helper.stack.get_visible_child() is det_label


@pytest.mark.unit
class TestFocusManagedListBox:
    """Test FocusManagedListBox focus management."""

    def test_focus_managed_listbox_initial_state(self) -> None:
        """Test FocusManagedListBox initializes with correct settings."""

        listbox = FocusManagedListBox()
        assert listbox.get_selection_mode() == Gtk.SelectionMode.NONE
        assert listbox.get_can_focus() is False

    def test_focus_managed_listbox_add_row_with_widget(self) -> None:
        """Test FocusManagedListBox tracks added rows and widgets."""

        listbox = FocusManagedListBox()
        row = Gtk.ListBoxRow()
        switch = Gtk.Switch()

        listbox.add_row_with_widget(row, switch)

        assert len(listbox._widgets) == 1
        assert len(listbox._rows) == 1
        assert listbox._widgets[0] is switch
        assert listbox._rows[0] is row

    def test_focus_managed_listbox_get_last_row(self) -> None:
        """Test FocusManagedListBox.get_last_row returns last added row."""

        listbox = FocusManagedListBox()
        assert listbox.get_last_row() is None

        row1 = Gtk.ListBoxRow()
        row2 = Gtk.ListBoxRow()
        listbox.add_row_with_widget(row1, Gtk.Switch())
        listbox.add_row_with_widget(row2, Gtk.Switch())

        assert listbox.get_last_row() is row2

    def test_focus_managed_listbox_multiple_rows(self) -> None:
        """Test FocusManagedListBox handles multiple rows correctly."""

        listbox = FocusManagedListBox()

        for _i in range(5):
            row = Gtk.ListBoxRow()
            switch = Gtk.Switch()
            listbox.add_row_with_widget(row, switch)

        assert len(listbox._widgets) == 5
        assert len(listbox._rows) == 5


@pytest.mark.unit
class TestPreferencesGridBase:
    """Test PreferencesGridBase UI helper methods."""

    def test_preferences_grid_base_initialization(self) -> None:
        """Test PreferencesGridBase initializes with correct properties."""

        grid = PreferencesGridBase("Test Tab")
        assert grid._tab_label == "Test Tab"
        assert grid._has_unsaved_changes is False
        assert grid.get_border_width() == 24  # pylint: disable=no-member

    def test_preferences_grid_base_get_label(self) -> None:
        """Test PreferencesGridBase.get_label returns Gtk.Label."""

        grid = PreferencesGridBase("My Tab")
        label = grid.get_label()

        assert isinstance(label, Gtk.Label)
        assert label.get_text() == "My Tab"

    def test_preferences_grid_base_has_changes(self) -> None:
        """Test PreferencesGridBase.has_changes tracks unsaved changes."""

        grid = PreferencesGridBase("Test")
        assert grid.has_changes() is False

        grid._has_unsaved_changes = True
        assert grid.has_changes() is True

    def test_create_frame_returns_frame_and_grid(self) -> None:
        """Test _create_frame returns Gtk.Frame and content Gtk.Grid."""

        grid = PreferencesGridBase("Test")
        frame, content = grid._create_frame("Section Label")

        assert isinstance(frame, Gtk.Frame)
        assert isinstance(content, Gtk.Grid)
        assert frame.get_child() is content

    def test_create_label_with_mnemonic(self) -> None:
        """Test _create_label creates label with mnemonic support."""

        grid = PreferencesGridBase("Test")
        label = grid._create_label("_Test Label")

        assert isinstance(label, Gtk.Label)
        # Label should use underline for mnemonic
        assert label.get_use_underline() is True

    def test_create_switch_row(self) -> None:
        """Test _create_switch_row creates row with label and switch."""

        grid = PreferencesGridBase("Test")
        handler_called = [False]

        def handler(_switch, _param):
            handler_called[0] = True

        row, switch, label = grid._create_switch_row("Enable Feature", handler, state=True)

        assert isinstance(row, Gtk.ListBoxRow)
        assert isinstance(switch, Gtk.Switch)
        assert isinstance(label, Gtk.Label)
        assert switch.get_active() is True

    def test_create_slider_row(self) -> None:
        """Test _create_slider_row creates row with label and scale."""

        grid = PreferencesGridBase("Test")
        adjustment = Gtk.Adjustment(
            value=50,
            lower=0,
            upper=100,
            step_increment=1,
            page_increment=10,
        )

        row, scale, label = grid._create_slider_row("Volume", adjustment)

        assert isinstance(row, Gtk.ListBoxRow)
        assert isinstance(scale, Gtk.Scale)
        assert isinstance(label, Gtk.Label)
        assert scale.get_value() == 50

    def test_create_info_listbox(self) -> None:
        """Test _create_info_listbox creates listbox with info message."""

        grid = PreferencesGridBase("Test")
        listbox = grid._create_info_listbox("This is an informational message.")

        assert isinstance(listbox, Gtk.ListBox)
        assert listbox.get_selection_mode() == Gtk.SelectionMode.NONE

    def test_create_scrolled_window(self) -> None:
        """Test _create_scrolled_window wraps widget in scrolled container."""

        grid = PreferencesGridBase("Test")
        label = Gtk.Label(label="Content")
        scrolled = grid._create_scrolled_window(label)

        assert isinstance(scrolled, Gtk.ScrolledWindow)
        assert scrolled.get_hexpand() is True
        assert scrolled.get_vexpand() is True


@pytest.mark.unit
class TestAutoPreferencesGrid:
    """Test AutoPreferencesGrid automatic UI building."""

    def _create_test_grid(self) -> tuple:
        """Create a test grid with various controls for testing."""

        values: dict[str, Any] = {
            "bool_setting": True,
            "int_setting": 5,
            "float_setting": 0.5,
            "enum_setting": "Option B",
        }

        def set_bool(x: bool) -> bool:
            values["bool_setting"] = x
            return True

        def set_int(x: int) -> bool:
            values["int_setting"] = x
            return True

        def set_float(x: float) -> bool:
            values["float_setting"] = x
            return True

        def set_enum(x: str) -> bool:
            values["enum_setting"] = x
            return True

        controls: list[ControlType] = [
            BooleanPreferenceControl(
                label="Enable Feature",
                getter=lambda: values["bool_setting"],
                setter=set_bool,
                prefs_key="boolSetting",
            ),
            IntRangePreferenceControl(
                label="Count",
                minimum=0,
                maximum=10,
                getter=lambda: values["int_setting"],
                setter=set_int,
                prefs_key="intSetting",
            ),
            FloatRangePreferenceControl(
                label="Rate",
                minimum=0.0,
                maximum=1.0,
                getter=lambda: values["float_setting"],
                setter=set_float,
                prefs_key="floatSetting",
            ),
            EnumPreferenceControl(
                label="Style",
                options=["Option A", "Option B", "Option C"],
                getter=lambda: values["enum_setting"],
                setter=set_enum,
                prefs_key="enumSetting",
            ),
        ]

        grid = AutoPreferencesGrid("Test Tab", controls)
        return grid, values, controls

    def test_auto_preferences_grid_builds_widgets(self) -> None:
        """Test AutoPreferencesGrid creates widgets for all controls."""

        grid, _values, controls = self._create_test_grid()

        assert len(grid._widgets) == len(controls)
        assert len(grid._controls) == len(controls)

    def test_auto_preferences_grid_boolean_control_creates_switch(self) -> None:
        """Test AutoPreferencesGrid creates Switch for boolean control."""

        grid, _values, _controls = self._create_test_grid()

        # First control is boolean, should create a Switch
        widget = grid.get_widget(0)
        assert isinstance(widget, Gtk.Switch)
        assert widget.get_active() is True

    def test_auto_preferences_grid_int_range_creates_spinbutton(self) -> None:
        """Test AutoPreferencesGrid creates SpinButton for int range control."""

        grid, _values, _controls = self._create_test_grid()

        widget = grid.get_widget(1)
        assert isinstance(widget, Gtk.SpinButton)
        assert int(widget.get_value()) == 5

    def test_auto_preferences_grid_float_range_creates_scale(self) -> None:
        """Test AutoPreferencesGrid creates Scale for float range control."""

        grid, _values, _controls = self._create_test_grid()

        widget = grid.get_widget(2)
        assert isinstance(widget, Gtk.Scale)
        assert widget.get_value() == pytest.approx(0.5, abs=0.01)

    def test_auto_preferences_grid_enum_creates_combobox(self) -> None:
        """Test AutoPreferencesGrid creates ComboBoxText for enum control."""

        grid, _values, _controls = self._create_test_grid()

        widget = grid.get_widget(3)
        assert isinstance(widget, Gtk.ComboBoxText)
        # Option B is at index 1
        assert widget.get_active() == 1

    def test_auto_preferences_grid_refresh_updates_widgets(self) -> None:
        """Test AutoPreferencesGrid.refresh updates widget values from getters."""

        current_value = [True]

        controls = [
            BooleanPreferenceControl(
                label="Test",
                getter=lambda: current_value[0],
                setter=lambda x: True,
            ),
        ]

        grid = AutoPreferencesGrid("Test", controls)
        switch = grid.get_widget(0)
        assert switch and switch.get_active()

        current_value[0] = False
        grid.refresh()
        assert switch and not switch.get_active()

    def test_auto_preferences_grid_save_settings_calls_setters(self) -> None:
        """Test AutoPreferencesGrid.save_settings calls all setters."""

        grid, values, _controls = self._create_test_grid()

        switch = grid.get_widget(0)
        switch.set_active(False)
        spin = grid.get_widget(1)
        spin.set_value(8)
        result = grid.save_settings()

        assert values["bool_setting"] is False
        assert values["int_setting"] == 8
        assert "boolSetting" in result
        assert result["boolSetting"] is False

    def test_auto_preferences_grid_save_returns_prefs_dict(self) -> None:
        """Test AutoPreferencesGrid.save_settings returns dict with prefs_keys."""

        grid, _values, _controls = self._create_test_grid()

        result = grid.save_settings()

        assert "boolSetting" in result
        assert "intSetting" in result
        assert "floatSetting" in result
        assert "enumSetting" in result

    def test_auto_preferences_grid_reload_clears_changes_flag(self) -> None:
        """Test AutoPreferencesGrid.reload clears unsaved changes flag."""

        grid, _values, _controls = self._create_test_grid()

        grid._has_unsaved_changes = True
        grid.reload()

        assert grid._has_unsaved_changes is False

    def test_auto_preferences_grid_get_widget_returns_none_for_invalid_index(self) -> None:
        """Test AutoPreferencesGrid.get_widget returns None for out-of-range index."""

        grid, _values, controls = self._create_test_grid()

        assert grid.get_widget(-1) is None
        assert grid.get_widget(len(controls)) is None
        assert grid.get_widget(100) is None

    def test_auto_preferences_grid_get_widget_for_control(self) -> None:
        """Test AutoPreferencesGrid.get_widget_for_control finds correct widget."""

        grid, _values, controls = self._create_test_grid()

        widget = grid.get_widget_for_control(controls[0])
        assert widget is not None
        assert isinstance(widget, Gtk.Switch)

        # Non-existent control should return None
        other_control = BooleanPreferenceControl(
            label="Other",
            getter=lambda: True,
            setter=lambda x: True,
        )
        assert grid.get_widget_for_control(other_control) is None

    def test_auto_preferences_grid_with_grouped_controls(self) -> None:
        """Test AutoPreferencesGrid groups controls by member_of field."""

        controls = [
            BooleanPreferenceControl(label="Ungrouped", getter=lambda: True, setter=lambda x: True),
            BooleanPreferenceControl(
                label="Group A Item 1",
                getter=lambda: True,
                setter=lambda x: True,
                member_of="Group A",
            ),
            BooleanPreferenceControl(
                label="Group A Item 2",
                getter=lambda: False,
                setter=lambda x: True,
                member_of="Group A",
            ),
            BooleanPreferenceControl(
                label="Group B Item",
                getter=lambda: True,
                setter=lambda x: True,
                member_of="Group B",
            ),
        ]

        grid = AutoPreferencesGrid("Test", controls)

        assert "Group A" in grid._group_labels
        assert "Group B" in grid._group_labels
        assert len(grid._widgets) == 4

    def test_auto_preferences_grid_sensitivity_callback(self) -> None:
        """Test AutoPreferencesGrid updates sensitivity based on callback."""

        primary_enabled = [True]

        def set_primary(x: bool) -> bool:
            primary_enabled[0] = x
            return True

        controls = [
            BooleanPreferenceControl(
                label="Primary",
                getter=lambda: primary_enabled[0],
                setter=set_primary,
                apply_immediately=True,
            ),
            BooleanPreferenceControl(
                label="Dependent",
                getter=lambda: True,
                setter=lambda x: True,
                determine_sensitivity=lambda: primary_enabled[0],
            ),
        ]

        grid = AutoPreferencesGrid("Test", controls)

        primary_switch = grid.get_widget(0)
        dependent_switch = grid.get_widget(1)

        assert dependent_switch and dependent_switch.get_sensitive()

        grid._initializing = False  # Allow change handlers to run
        assert primary_switch
        primary_switch.set_active(False)

        assert dependent_switch and not dependent_switch.get_sensitive()

    def test_auto_preferences_grid_with_info_message(self) -> None:
        """Test AutoPreferencesGrid displays info message when provided."""

        controls = [
            BooleanPreferenceControl(label="Test", getter=lambda: True, setter=lambda x: True),
        ]

        grid = AutoPreferencesGrid("Test", controls, info_message="This is helpful information.")

        assert grid._info_listbox is not None


@pytest.mark.unit
class TestAutoPreferencesGridSelectionControl:
    """Test AutoPreferencesGrid with SelectionPreferenceControl."""

    def test_selection_control_creates_combobox_without_actions(self) -> None:
        """Test SelectionPreferenceControl without actions creates ComboBoxText."""

        controls = [
            SelectionPreferenceControl(
                label="Choose",
                options=["A", "B", "C"],
                getter=lambda: "B",
                setter=lambda x: True,
            ),
        ]

        grid = AutoPreferencesGrid("Test", controls)
        widget = grid.get_widget(0)

        assert isinstance(widget, Gtk.ComboBoxText)
        assert widget.get_active() == 1

    def test_selection_control_creates_radio_buttons_with_actions(self) -> None:
        """Test SelectionPreferenceControl with actions creates RadioButtons."""

        def get_actions(_value):
            return [("Edit", "edit-symbolic", lambda: None)]

        controls = [
            SelectionPreferenceControl(
                label="Profile",
                options=["Default", "Custom"],
                getter=lambda: "Default",
                setter=lambda x: True,
                get_actions_for_option=get_actions,
            ),
        ]

        grid = AutoPreferencesGrid("Test", controls)
        widget = grid.get_widget(0)

        assert isinstance(widget, RadioButtonWithActions)

    def test_selection_control_with_values_mapping(self) -> None:
        """Test SelectionPreferenceControl maps display options to values."""

        current_value = [100]

        def set_level(x: int) -> bool:
            current_value[0] = x
            return True

        controls = [
            SelectionPreferenceControl(
                label="Level",
                options=["Low", "Medium", "High"],
                values=[50, 100, 150],
                getter=lambda: current_value[0],
                setter=set_level,
                prefs_key="level",
            ),
        ]

        grid = AutoPreferencesGrid("Test", controls)
        widget = grid.get_widget(0)

        assert widget and widget.get_active() == 1

        widget.set_active(2)
        result = grid.save_settings()

        assert current_value[0] == 150
        assert result["level"] == 150


@pytest.mark.unit
class TestPreferencesGridBaseMultiPageStack:
    """Test PreferencesGridBase multi-page stack functionality."""

    def test_create_multi_page_stack_basic(self) -> None:
        """Test _create_multi_page_stack creates stack with categories."""

        class TestGrid(PreferencesGridBase):
            """Test grid."""

            def __init__(self):
                super().__init__("Test")
                self._initializing = True

                child1 = PreferencesGridBase("Child 1")
                child2 = PreferencesGridBase("Child 2")

                enable_listbox, stack, categories = self._create_multi_page_stack(
                    enable_label="Enable",
                    enable_getter=lambda: True,
                    enable_setter=lambda x: True,
                    categories=[
                        ("General", "general", child1),
                        ("Advanced", "advanced", child2),
                    ],
                    main_title="Test Settings",
                )

                self.my_enable_listbox = enable_listbox
                self.my_stack = stack
                self.my_categories = categories
                self._initializing = False

        grid = TestGrid()

        assert grid.my_stack is not None
        assert grid.my_categories is not None
        assert grid.my_enable_listbox is not None

    def test_create_multi_page_stack_without_enable(self) -> None:
        """Test _create_multi_page_stack without enable switch."""

        class TestGrid(PreferencesGridBase):
            """Test grid."""

            def __init__(self):
                super().__init__("Test")
                self._initializing = True

                child = PreferencesGridBase("Child")

                enable_listbox, stack, _categories = self._create_multi_page_stack(
                    enable_label=None,
                    enable_getter=None,
                    enable_setter=None,
                    categories=[("General", "general", child)],
                    main_title="Test",
                )

                self.my_enable_listbox = enable_listbox
                self.my_stack = stack
                self._initializing = False

        grid = TestGrid()

        assert grid.my_enable_listbox is None
        assert grid.my_stack is not None


@pytest.mark.unit
class TestPreferencesGridBaseStackedPreferences:
    """Test PreferencesGridBase stacked drill-down preferences."""

    def test_create_stacked_preferences(self) -> None:
        """Test _create_stacked_preferences creates stack with categories and detail."""

        class TestGrid(PreferencesGridBase):
            """Test grid."""

            def __init__(self):
                super().__init__("Test")
                self.activated_row = None

                stack, categories, detail = self._create_stacked_preferences(
                    on_category_activated=self._on_category,
                    on_detail_row_activated=None,
                )
                self.my_stack = stack
                self.my_categories = categories
                self.my_detail = detail

            def _on_category(self, row):
                self.activated_row = row

        grid = TestGrid()

        assert grid.my_stack is not None
        assert grid.my_categories is not None
        assert grid.my_detail is not None
        assert isinstance(grid.my_stack, Gtk.Stack)
        assert isinstance(grid.my_categories, Gtk.ListBox)
        assert isinstance(grid.my_detail, Gtk.ListBox)

    def test_add_stack_category_row(self) -> None:
        """Test _add_stack_category_row adds category with correct properties."""

        class TestGrid(PreferencesGridBase):
            """Test grid."""

            def __init__(self):
                super().__init__("Test")
                _stack, categories, _detail = self._create_stacked_preferences(
                    on_category_activated=lambda r: None,
                )
                self.my_categories = categories

        grid = TestGrid()

        row = grid._add_stack_category_row(grid.my_categories, "Speech Settings", category="speech")

        assert isinstance(row, CategoryListBoxRow)
        assert row.category == "speech"

    def test_add_stack_category_row_chevron_accessible_role(self) -> None:
        """Test _add_stack_category_row chevron has PUSH_BUTTON accessible role."""

        gi.require_version("Atk", "1.0")
        from gi.repository import Atk

        class TestGrid(PreferencesGridBase):
            """Test grid."""

            def __init__(self):
                super().__init__("Test")
                _stack, categories, _detail = self._create_stacked_preferences(
                    on_category_activated=lambda r: None,
                )
                self.my_categories = categories

        grid = TestGrid()
        row = grid._add_stack_category_row(grid.my_categories, "Test Category", category="test")

        # Find the chevron image in the row's hbox
        hbox = row.get_child()
        chevron = None
        for child in hbox.get_children():
            if isinstance(child, Gtk.Image):
                chevron = child
                break

        assert chevron is not None, "Chevron image not found in category row"

        chevron_accessible = chevron.get_accessible()
        assert chevron_accessible is not None
        assert chevron_accessible.get_role() == Atk.Role.BUTTON
        assert chevron_accessible.get_name() == ""

    def test_show_stack_categories_and_detail(self) -> None:
        """Test _show_stack_categories and _show_stack_detail switch views."""

        class TestGrid(PreferencesGridBase):
            """Test grid."""

            def __init__(self):
                super().__init__("Test")
                stack, categories, detail = self._create_stacked_preferences(
                    on_category_activated=lambda r: None,
                )
                self.my_stack = stack
                self.my_categories = categories
                self.my_detail = detail

        grid = TestGrid()

        grid.my_stack.show_all()  # pylint: disable=no-member

        cat_scrolled = grid.my_stack.get_child_by_name("categories")
        det_scrolled = grid.my_stack.get_child_by_name("detail")

        grid._show_stack_detail()
        assert grid.my_stack.get_visible_child() is det_scrolled

        grid._show_stack_categories()
        assert grid.my_stack.get_visible_child() is cat_scrolled



# Unit tests for pronunciation_dictionary_manager.py.
#
# Copyright 2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=wrong-import-position
# pylint: disable=protected-access
# pylint: disable=import-outside-toplevel

"""Unit tests for pronunciation_dictionary_manager.py."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from .orca_test_context import OrcaTestContext


@pytest.mark.unit
class TestPronunciationDictionaryManager:
    """Test PronunciationDictionaryManager class."""

    def _setup_manager(self, test_context: OrcaTestContext):
        """Set up mocks for pronunciation_dictionary_manager dependencies."""

        additional_modules = [
            "orca.guilabels",
            "orca.messages",
            "orca.preferences_grid_base",
            "orca.script_manager",
            "orca.speech_manager",
        ]
        essential_modules = test_context.setup_shared_dependencies(additional_modules)

        # Set up guilabels with required constants
        guilabels = essential_modules["orca.guilabels"]
        guilabels.PRONUNCIATION = "Pronunciation"
        guilabels.PRONUNCIATION_DICTIONARY = "Pronunciation Dictionary"
        guilabels.PRONUNCIATION_DICTIONARY_INFO = "Add custom pronunciations."
        guilabels.DICTIONARY_NEW_ENTRY = "New Entry"
        guilabels.DICTIONARY_DELETE = "Delete"
        guilabels.DICTIONARY_EMPTY = "No entries"
        guilabels.DICTIONARY_ACTUAL_STRING = "Actual String"
        guilabels.DICTIONARY_REPLACEMENT_STRING = "Replacement String"
        guilabels.ADD_NEW_PRONUNCIATION = "Add New Pronunciation"
        guilabels.EDIT_PRONUNCIATION = "Edit Pronunciation"
        guilabels.DIALOG_CANCEL = "Cancel"
        guilabels.DIALOG_ADD = "Add"
        guilabels.DIALOG_EDIT = "Edit"

        # Set up messages
        messages = essential_modules["orca.messages"]
        messages.PRONUNCIATION_DELETED = "Pronunciation %s deleted"

        # Set up speech_manager
        speech_manager_mock = essential_modules["orca.speech_manager"]
        speech_instance = speech_manager_mock.get_manager.return_value
        speech_instance.get_use_pronunciation_dictionary.return_value = True
        speech_instance.set_use_pronunciation_dictionary.return_value = True

        # Import and return the module
        from orca import pronunciation_dictionary_manager

        return pronunciation_dictionary_manager, essential_modules

    def test_get_manager_returns_singleton(self, test_context: OrcaTestContext) -> None:
        """Test get_manager returns the same instance."""

        module, _mocks = self._setup_manager(test_context)
        manager1 = module.get_manager()
        manager2 = module.get_manager()

        assert manager1 is manager2

    def test_manager_initial_state(self, test_context: OrcaTestContext) -> None:
        """Test manager initializes with empty dictionary."""

        module, _mocks = self._setup_manager(test_context)
        manager = module.get_manager()

        # Reset the dictionary to ensure clean state
        manager.set_dictionary({})

        assert manager.get_dictionary() == {}

    def test_get_pronunciation_returns_word_when_not_found(
        self,
        test_context: OrcaTestContext,
    ) -> None:
        """Test get_pronunciation returns the original word when not in dictionary."""

        module, _mocks = self._setup_manager(test_context)
        manager = module.get_manager()
        manager.set_dictionary({})

        result = manager.get_pronunciation("hello")

        assert result == "hello"

    def test_get_pronunciation_returns_replacement_when_found(
        self,
        test_context: OrcaTestContext,
    ) -> None:
        """Test get_pronunciation returns replacement when word is in dictionary."""

        module, _mocks = self._setup_manager(test_context)
        manager = module.get_manager()
        manager.set_dictionary({})

        manager.set_pronunciation("hello", "hi there")
        result = manager.get_pronunciation("hello")

        assert result == "hi there"

    def test_get_pronunciation_is_case_insensitive(self, test_context: OrcaTestContext) -> None:
        """Test get_pronunciation lookup is case insensitive."""

        module, _mocks = self._setup_manager(test_context)
        manager = module.get_manager()
        manager.set_dictionary({})

        manager.set_pronunciation("HELLO", "hi there")

        assert manager.get_pronunciation("hello") == "hi there"
        assert manager.get_pronunciation("Hello") == "hi there"
        assert manager.get_pronunciation("HELLO") == "hi there"

    def test_set_pronunciation_stores_lowercase_key(self, test_context: OrcaTestContext) -> None:
        """Test set_pronunciation stores key as lowercase."""

        module, _mocks = self._setup_manager(test_context)
        manager = module.get_manager()
        manager.set_dictionary({})

        manager.set_pronunciation("MixedCase", "replacement")

        dictionary = manager.get_dictionary()
        assert "mixedcase" in dictionary
        assert "MixedCase" not in dictionary

    def test_set_dictionary_replaces_entire_dictionary(self, test_context: OrcaTestContext) -> None:
        """Test set_dictionary replaces the entire dictionary."""

        module, _mocks = self._setup_manager(test_context)
        manager = module.get_manager()

        manager.set_pronunciation("old", "old replacement")
        manager.set_dictionary({"new": "new replacement"})

        dictionary = manager.get_dictionary()
        assert "old" not in dictionary
        assert dictionary.get("new") == "new replacement"

    def test_multiple_pronunciations(self, test_context: OrcaTestContext) -> None:
        """Test manager handles multiple pronunciations correctly."""

        module, _mocks = self._setup_manager(test_context)
        manager = module.get_manager()
        manager.set_dictionary({})

        manager.set_pronunciation("word1", "replacement1")
        manager.set_pronunciation("word2", "replacement2")
        manager.set_pronunciation("word3", "replacement3")

        assert manager.get_pronunciation("word1") == "replacement1"
        assert manager.get_pronunciation("word2") == "replacement2"
        assert manager.get_pronunciation("word3") == "replacement3"
        assert len(manager.get_dictionary()) == 3

    def test_set_pronunciation_overwrites_existing(self, test_context: OrcaTestContext) -> None:
        """Test set_pronunciation overwrites existing entry for same word."""

        module, _mocks = self._setup_manager(test_context)
        manager = module.get_manager()
        manager.set_dictionary({})

        manager.set_pronunciation("word", "first replacement")
        manager.set_pronunciation("word", "second replacement")

        assert manager.get_pronunciation("word") == "second replacement"
        assert len(manager.get_dictionary()) == 1


@pytest.mark.unit
class TestPronunciationDictionaryManagerIntegration:
    """Integration tests for pronunciation dictionary manager."""

    def _setup_manager(self, test_context: OrcaTestContext):
        """Set up mocks for pronunciation_dictionary_manager dependencies."""

        additional_modules = [
            "orca.guilabels",
            "orca.messages",
            "orca.preferences_grid_base",
            "orca.script_manager",
            "orca.speech_manager",
        ]
        essential_modules = test_context.setup_shared_dependencies(additional_modules)

        # Set up guilabels with required constants
        guilabels = essential_modules["orca.guilabels"]
        guilabels.PRONUNCIATION = "Pronunciation"
        guilabels.PRONUNCIATION_DICTIONARY = "Pronunciation Dictionary"
        guilabels.PRONUNCIATION_DICTIONARY_INFO = "Add custom pronunciations."
        guilabels.DICTIONARY_NEW_ENTRY = "New Entry"
        guilabels.DICTIONARY_DELETE = "Delete"
        guilabels.DICTIONARY_EMPTY = "No entries"
        guilabels.DICTIONARY_ACTUAL_STRING = "Actual String"
        guilabels.DICTIONARY_REPLACEMENT_STRING = "Replacement String"
        guilabels.ADD_NEW_PRONUNCIATION = "Add New Pronunciation"
        guilabels.EDIT_PRONUNCIATION = "Edit Pronunciation"
        guilabels.DIALOG_CANCEL = "Cancel"
        guilabels.DIALOG_ADD = "Add"
        guilabels.DIALOG_EDIT = "Edit"

        # Set up messages
        messages = essential_modules["orca.messages"]
        messages.PRONUNCIATION_DELETED = "Pronunciation %s deleted"

        # Set up speech_manager
        speech_manager_mock = essential_modules["orca.speech_manager"]
        speech_instance = speech_manager_mock.get_manager.return_value
        speech_instance.get_use_pronunciation_dictionary.return_value = True
        speech_instance.set_use_pronunciation_dictionary.return_value = True

        # Import and return the module
        from orca import pronunciation_dictionary_manager

        return pronunciation_dictionary_manager, essential_modules

    def test_pronunciation_application_to_text(self, test_context: OrcaTestContext) -> None:
        """Test that pronunciation dictionary can be used to process text."""

        module, _mocks = self._setup_manager(test_context)
        manager = module.get_manager()
        manager.set_dictionary({})

        manager.set_pronunciation("github", "git hub")
        manager.set_pronunciation("cli", "command line interface")

        # Simulate applying pronunciations to a sentence
        words = ["Check", " ", "out", " ", "the", " ", "GitHub", " ", "CLI"]
        result = "".join(map(manager.get_pronunciation, words))

        # GitHub becomes "git hub" because lookup is case insensitive
        assert "git hub" in result
        # CLI becomes "command line interface"
        assert "command line interface" in result

    def test_empty_dictionary_returns_original_words(self, test_context: OrcaTestContext) -> None:
        """Test that empty dictionary returns all original words unchanged."""

        module, _mocks = self._setup_manager(test_context)
        manager = module.get_manager()
        manager.set_dictionary({})

        words = ["Hello", " ", "world", "!"]
        result = "".join(map(manager.get_pronunciation, words))

        assert result == "Hello world!"
